"""Dummy_cursor module."""


from pineboolib.core.utils import logging
from pineboolib.application.metadata import pntablemetadata, pnaction
from pineboolib.application.qsatypes import date
from pineboolib.application.database import utils
from pineboolib import application
import datetime


from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.database.orm import basemodel  # pragma: no cover
    from pineboolib.interfaces import iconnection  # pragma: no cover
    from pineboolib.application.database import pnsqlcursor  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class DummyCursor(object):
    """DummyCursor class."""

    Insert: int = 0
    Edit: int = 1
    Del: int = 2
    Browse: int = 3

    _parent: "basemodel.BaseModel"

    def __init__(self, parent_model: "basemodel.BaseModel") -> None:
        """Initialize."""

        self._parent = parent_model

    def modeAccess(self) -> int:
        """Return mode_access."""

        return (
            self._parent.mode_access
            if self._parent._current_mode is None
            else self._parent._current_mode
        )

    def valueBuffer(self, field_name: str, return_none: bool = False) -> Any:
        """Return field value."""

        value = (
            self._parent._cached_bufferchanged[field_name]
            if field_name in self._parent._cached_bufferchanged.keys()
            else getattr(self._parent, field_name, None)
        )
        meta_table = self._parent.table_metadata()
        meta_field = meta_table.field(field_name)
        if meta_field is None:
            LOGGER.warning(
                "dummy_cursor.valueBuffer. Field metadata %s not found in %s table."
                % (field_name, meta_table.name())
            )
        else:
            type_ = meta_field.type()  # type: ignore [union-attr]
            if value:
                value = utils.resolve_qsa_value(type_, value)
            elif not return_none:
                value = utils.resolve_empty_qsa_value(type_)

        return value

    def valueBufferCopy(self, field_name: str, return_none: bool = False) -> Any:
        """Return field value copy."""

        value = getattr(self._parent.copy(), field_name)

        meta_table = self._parent.table_metadata()
        meta_field = meta_table.field(field_name)
        if meta_field is None:
            LOGGER.warning(
                "dummy_cursor.valueBufferCopy. Field metadata %s not found in %s table."
                % (field_name, meta_table.name())
            )
        else:
            type_ = meta_field.type()  # type: ignore [union-attr]
            if value:
                value = utils.resolve_qsa_value(type_, value)
            elif not return_none:
                value = utils.resolve_empty_qsa_value(type_)

        return value

    def setValueBuffer(self, field_name: str, value: Any) -> Any:
        """Set field value."""

        meta_table = self._parent.table_metadata()
        meta_field = meta_table.field(field_name)
        if meta_field is None:
            LOGGER.warning(
                "dummy_cursor.setValueBuffer. Field metadata %s not found in %s table."
                % (field_name, meta_table.name())
            )
        else:
            type_ = meta_field.type()  # type: ignore [union-attr]

            if type_ == "date":
                if not isinstance(value, datetime.date):
                    value = datetime.datetime.strptime(str(value)[0:10], "%Y-%m-%d").date()
            elif type_ == "time":
                if not isinstance(value, datetime.time):
                    value = datetime.datetime.strptime(str(value)[:8], "%H:%M:%S").time()

        setattr(self._parent, field_name, value)

    def setValueBufferCopy(self, field_name: str, value: Any) -> Any:
        """Set field value."""

        meta_table = self._parent.table_metadata()
        meta_field = meta_table.field(field_name)
        if meta_field is None:
            LOGGER.warning(
                "dummy_cursor.setValueBufferCopy. Field metadata %s not found in %s table."
                % (field_name, meta_table.name())
            )
        else:
            type_ = meta_field.type()  # type: ignore [union-attr]

            if type_ == "date":
                if isinstance(value, date.Date):
                    value = datetime.datetime.strptime(str(value)[0:10], "%Y-%m-%d").date()

        setattr(self._parent.copy(), field_name, value)

    def isNull(self, field_name: str) -> bool:
        """Return if value is Null."""

        return getattr(self._parent, field_name) is None

    def isValid(self):
        """Return if cursor is valid."""

        return self._parent is not None

    def setNull(self, field_name: str):
        """Set value to Null."""

        setattr(self._parent, field_name, None)

    def metadata(self) -> "pntablemetadata.PNTableMetaData":
        """Return metadata."""

        return self._parent.table_metadata()

    def db(self) -> "iconnection.IConnection":
        """Return pnconnection."""

        return application.PROJECT.conn_manager.useConn(
            self._parent._session._conn_name  # type: ignore [union-attr] # noqa: F821
        )

    def table(self) -> str:
        """Return table name."""

        return self._parent.__tablename__  # type: ignore [attr-defined] # noqa: F821

    def primaryKey(self) -> str:
        """Return primary key name."""

        return self._parent.pk_name

    def cursorRelation(self) -> Optional["pnsqlcursor.PNSqlCursor"]:
        """Return cursor Relation."""

        LOGGER.warning("FIXME: Avoid using cursorRelation from a DummyCursor returns empty always")
        return None

    def get_bc_signal(self):
        """Return beforeCommit fake signal."""

        return self._parent.bufferChanged

    def isModifiedBuffer(self):
        """Return if buffer is modified."""

        return len(self._parent.changes()) > 0

    def action(self) -> "pnaction.PNAction":
        """Return PNAction."""

        return pnaction.PNAction(self._parent._action)  # type: ignore [arg-type]

    def getattr(self, name: str) -> None:
        """Search unknown functions."""

        raise Exception("PLEASE IMPLEMENT DummyCursor.%s." % name)

    bufferChanged = property(get_bc_signal)



"""Qpushbutton module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtCore  # type: ignore[import]
from pineboolib.core import decorators

from typing import Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtGui  # pragma: no cover


class QPushButton(QtWidgets.QPushButton):
    """QPushButton class."""

    def __init__(self, *args, **kwargs) -> None:
        """Inicialize."""
        super(QPushButton, self).__init__(*args, **kwargs)

    def setTextLabel(self, text: str) -> None:
        """Set text label."""
        self.setText(text)

    @decorators.not_implemented_warn
    def setTextPosition(self, pos: int) -> None:
        """Set text position."""
        pass

    @decorators.not_implemented_warn
    def setUsesBigPixmap(self, uses_big: bool) -> None:
        """Set use big pixmap."""

        pass

    @decorators.not_implemented_warn
    def setUsesTextLabel(self, uses_text: bool) -> None:
        """Set use text label."""
        pass

    @property
    def pixmap(self) -> "QtGui.QIcon":
        """Return pixmap."""

        return self.icon()

    @pixmap.setter
    def pixmap(self, value: "QtGui.QIcon") -> None:
        """Set pixmap."""
        return self.setIcon(value)

    def setPixmap(self, value: "QtGui.QIcon") -> None:
        """Set pixmap."""
        return self.setIcon(value)

    def getToggleButton(self) -> bool:
        """Return if is checkable."""

        return self.isCheckable()

    def setToggleButton(self, toggle: bool) -> None:
        """Set if is checkable."""

        self.setCheckable(toggle)

    def getOn(self) -> bool:
        """Return if is checked."""
        return self.isChecked()

    def setOn(self, value: bool) -> None:
        """Set if is checked."""

        self.setChecked(value)

    def getText(self) -> str:
        """Return text."""

        return super().text()

    def setText(self, val: Optional[str]) -> None:
        """Set text."""

        if self.maximumWidth() < 33 and len(val) > 4:  # type: ignore [arg-type]
            val = ""
        super().setText(val)  # type: ignore [arg-type]

    def setMaximumSize(
        self, width: Union[int, QtCore.QSize] = 30, height: Optional[int] = None
    ) -> None:
        """Set Maximun size."""

        if isinstance(width, int):
            if height is None:
                height = width

            super().setMaximumSize(width, height)
        else:
            super().setMaximumSize(width)

    def isEnabled(self) -> bool:
        """Return if the button is enabled. Overloaded by property assign."""
        return super().isEnabled()

    def setEnabled(self, enabled: bool) -> None:
        """Set if the button is enabled. Overloaded by property assign."""
        super().setEnabled(enabled)

    toggleButton: bool = property(  # type: ignore [assignment] # noqa: F821
        getToggleButton, setToggleButton
    )
    on: bool = property(getOn, setOn)  # type: ignore [assignment] # noqa: F821
    text: str = property(getText, setText)  # type: ignore [assignment] # noqa: F821
    enabled: bool = property(isEnabled, setEnabled)  # type: ignore [assignment] # noqa: F821



"""Test_acls module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.fllegacy import systype
from pineboolib.core import settings
from pineboolib.core.utils import utils_base
from pineboolib.application.acls import pnaccesscontrollists
from pineboolib import application, qsa
from pineboolib.application.acls.tests import fixture_path


class TestACLS(unittest.TestCase):
    """TestPNBuffer Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()
        utils_base.FORCE_DESKTOP = True
        application.PROJECT.db_admin_mode = True

        # Install package
        qsa_sys = qsa.qsa.sys
        path = fixture_path("controlacceso.eneboopkg")
        qsa_sys.loadModules(path, False)

        # install acls
        from pineboolib.application.database import pnsqlcursor

        cursor_flgroups = pnsqlcursor.PNSqlCursor("flgroups")
        cursor_flgroups.setModeAccess(cursor_flgroups.Insert)
        cursor_flgroups.refreshBuffer()
        cursor_flgroups.setValueBuffer("idgroup", "usuarios")
        cursor_flgroups.setValueBuffer("descripcion", "Grupo usuarios")
        cursor_flgroups.commitBuffer()

        cursor_flusers = pnsqlcursor.PNSqlCursor("flusers")
        cursor_flusers.setModeAccess(cursor_flusers.Insert)
        cursor_flusers.refreshBuffer()
        cursor_flusers.setValueBuffer("iduser", "memory_user")
        cursor_flusers.setValueBuffer("idgroup", "usuarios")
        cursor_flusers.setValueBuffer("descripcion", "test user")
        cursor_flusers.commitBuffer()

        cursor_flacls = pnsqlcursor.PNSqlCursor("flacls")
        cursor_flacls.setModeAccess(cursor_flacls.Insert)
        cursor_flacls.setActivatedCheckIntegrity(False)
        cursor_flacls.refreshBuffer()
        cursor_flacls.setValueBuffer("idacl", "primera")
        cursor_flacls.setValueBuffer("descripcion", "first acl")
        cursor_flacls.setValueBuffer("prioridadgrupointro", 2)
        cursor_flacls.setValueBuffer("prioridadusuariointro", 1)
        cursor_flacls.commitBuffer()
        cursor_flacls.setModeAccess(cursor_flacls.Insert)
        cursor_flacls.setActivatedCheckIntegrity(False)
        cursor_flacls.refreshBuffer()
        cursor_flacls.setValueBuffer("idacl", "segunda")
        cursor_flacls.setValueBuffer("descripcion", "second acl")
        cursor_flacls.setValueBuffer("prioridadgrupointro", 2)
        cursor_flacls.setValueBuffer("prioridadusuariointro", 1)
        cursor_flacls.commitBuffer()
        cursor_flacls.setModeAccess(cursor_flacls.Insert)
        cursor_flacls.setActivatedCheckIntegrity(False)
        cursor_flacls.refreshBuffer()
        cursor_flacls.setValueBuffer("idacl", "tercera")
        cursor_flacls.setValueBuffer("descripcion", "third acl")
        cursor_flacls.setValueBuffer("prioridadgrupointro", 2)
        cursor_flacls.setValueBuffer("prioridadusuariointro", 1)
        cursor_flacls.commitBuffer()
        cursor_flacls.setModeAccess(cursor_flacls.Insert)
        cursor_flacls.setActivatedCheckIntegrity(False)
        cursor_flacls.refreshBuffer()
        cursor_flacls.setValueBuffer("idacl", "cuarta")
        cursor_flacls.setValueBuffer("descripcion", "fourth acl")
        cursor_flacls.setValueBuffer("prioridadgrupointro", 2)
        cursor_flacls.setValueBuffer("prioridadusuariointro", 1)
        cursor_flacls.commitBuffer()
        cursor_flacls.setModeAccess(cursor_flacls.Insert)
        cursor_flacls.setActivatedCheckIntegrity(False)
        cursor_flacls.refreshBuffer()
        cursor_flacls.setValueBuffer("idacl", "final")
        cursor_flacls.setValueBuffer("descripcion", "clear")
        cursor_flacls.setValueBuffer("prioridadgrupointro", 2)
        cursor_flacls.setValueBuffer("prioridadusuariointro", 1)
        cursor_flacls.commitBuffer()

        cursor_flacs = pnsqlcursor.PNSqlCursor("flacs")
        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 1)
        cursor_flacs.setValueBuffer("tipo", "table")
        cursor_flacs.setValueBuffer("nombre", "flgroups")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "primera")
        cursor_flacs.setValueBuffer("descripcion", "Sistema:Administración:table:Maestro:flgroup")
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")
        cursor_flacs.commitBuffer()
        id_acs_1 = cursor_flacs.valueBuffer("idac")

        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 2)
        cursor_flacs.setValueBuffer("tipo", "table")
        cursor_flacs.setValueBuffer("nombre", "flmodules")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "primera")
        cursor_flacs.setValueBuffer("descripcion", "Sistema:Administración:table:Maestro:flmodules")
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")
        cursor_flacs.commitBuffer()
        id_acs_2 = cursor_flacs.valueBuffer("idac")

        # global tables '--'
        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 3)
        cursor_flacs.setValueBuffer("tipo", "table")
        cursor_flacs.setValueBuffer("nombre", "flareas")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "segunda")
        cursor_flacs.setValueBuffer("descripcion", "Sistema:Administración:table:Maestro:flareas")
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")
        cursor_flacs.setValueBuffer("permiso", "--")
        cursor_flacs.commitBuffer()

        # global tables 'r-'
        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 4)
        cursor_flacs.setValueBuffer("tipo", "table")
        cursor_flacs.setValueBuffer("nombre", "flusers")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "segunda")
        cursor_flacs.setValueBuffer("descripcion", "Sistema:Administración:table:Maestro:flusers")
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")
        cursor_flacs.setValueBuffer("permiso", "r-")
        cursor_flacs.commitBuffer()

        # global tables 'rw'
        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 5)
        cursor_flacs.setValueBuffer("tipo", "table")
        cursor_flacs.setValueBuffer("nombre", "fltest")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "segunda")
        cursor_flacs.setValueBuffer("descripcion", "Sistema:Administración:table:Maestro:fltest")
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")
        cursor_flacs.setValueBuffer("permiso", "rw")
        cursor_flacs.commitBuffer()

        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 6)
        cursor_flacs.setValueBuffer("tipo", "form")
        cursor_flacs.setValueBuffer("nombre", "formRecordflmodules")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "tercera")
        cursor_flacs.setValueBuffer(
            "descripcion", "Sistema:Administración:form:Edición:formRecordflmodules"
        )
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Edición")
        cursor_flacs.commitBuffer()

        id_acs_3 = cursor_flacs.valueBuffer("idac")

        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 7)
        cursor_flacs.setValueBuffer("tipo", "mainwindow")
        cursor_flacs.setValueBuffer("nombre", "sys")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("idacl", "cuarta")
        cursor_flacs.setValueBuffer("descripcion", "Sistema:Administración:mainwindow:Maestro:sys")
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")
        cursor_flacs.setValueBuffer("permiso", "--")
        cursor_flacs.commitBuffer()

        id_acs_4 = cursor_flacs.valueBuffer("idac")

        # global form 'r-'
        cursor_flacs.setModeAccess(cursor_flacs.Insert)
        cursor_flacs.setActivatedCheckIntegrity(False)
        cursor_flacs.refreshBuffer()
        cursor_flacs.setValueBuffer("prioridad", 8)
        cursor_flacs.setValueBuffer("tipo", "form")
        cursor_flacs.setValueBuffer("nombre", "formflmodules")
        cursor_flacs.setValueBuffer("idgroup", "usuarios")
        cursor_flacs.setValueBuffer("permiso", "r-")
        cursor_flacs.setValueBuffer("idacl", "tercera")
        cursor_flacs.setValueBuffer(
            "descripcion", "Sistema:Administración:form:Maestro:formflmodules"
        )
        cursor_flacs.setValueBuffer("degrupo", True)
        cursor_flacs.setValueBuffer("idarea", "sys")
        cursor_flacs.setValueBuffer("idmodule", "sys")
        cursor_flacs.setValueBuffer("tipoform", "Maestro")

        cursor_flacs.commitBuffer()

        cursor_flacos = pnsqlcursor.PNSqlCursor("flacos")
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        # single table 'r-'
        cursor_flacos.setValueBuffer("nombre", "descripcion")
        cursor_flacos.setValueBuffer("permiso", "r-")
        cursor_flacos.setValueBuffer("idac", id_acs_1)
        cursor_flacos.setValueBuffer("tipocontrol", "Tabla")
        cursor_flacos.commitBuffer()
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        # single table '--'
        cursor_flacos.setValueBuffer("nombre", "idgroup")
        cursor_flacos.setValueBuffer("permiso", "--")
        cursor_flacos.setValueBuffer("idac", id_acs_1)
        cursor_flacos.setValueBuffer("tipocontrol", "Tabla")
        cursor_flacos.commitBuffer()
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        # single table 'rw'
        cursor_flacos.setValueBuffer("nombre", "descripcion")
        cursor_flacos.setValueBuffer("permiso", "rw")
        cursor_flacos.setValueBuffer("idac", id_acs_2)
        cursor_flacos.commitBuffer()

        # single form 'r-'
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        cursor_flacos.setValueBuffer("nombre", "botonExportar")
        cursor_flacos.setValueBuffer("descripcion", "Botón:botonExportar")
        cursor_flacos.setValueBuffer("tipocontrol", "Botón")
        cursor_flacos.setValueBuffer("permiso", "r-")
        cursor_flacos.setValueBuffer("idac", id_acs_3)
        cursor_flacos.commitBuffer()

        # single form '--'
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        cursor_flacos.setValueBuffer("nombre", "botonCargar")
        cursor_flacos.setValueBuffer("descripcion", "Botón:botonCargar")
        cursor_flacos.setValueBuffer("tipocontrol", "Botón")
        cursor_flacos.setValueBuffer("permiso", "--")
        cursor_flacos.setValueBuffer("idac", id_acs_3)
        cursor_flacos.commitBuffer()

        # single mainwindow '--'
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        cursor_flacos.setValueBuffer("nombre", "ebcomportamiento")
        cursor_flacos.setValueBuffer("descripcion", "Todos:ebcomportamiento")
        cursor_flacos.setValueBuffer("tipocontrol", "Todos")
        cursor_flacos.setValueBuffer("permiso", "--")
        cursor_flacos.setValueBuffer("idac", id_acs_4)
        cursor_flacos.commitBuffer()

        # single mainwindow '-w'
        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        cursor_flacos.setValueBuffer("nombre", "flgroups")
        cursor_flacos.setValueBuffer("descripcion", "Todos:flgroups")
        cursor_flacos.setValueBuffer("tipocontrol", "Todos")
        cursor_flacos.setValueBuffer("permiso", "-w")
        cursor_flacos.setValueBuffer("idac", id_acs_4)
        cursor_flacos.commitBuffer()

        cursor_flacos.setModeAccess(cursor_flacos.Insert)
        cursor_flacos.refreshBuffer()
        cursor_flacos.setValueBuffer("nombre", "flareas")
        cursor_flacos.setValueBuffer("descripcion", "Todos:flareas")
        cursor_flacos.setValueBuffer("tipocontrol", "Todos")
        cursor_flacos.setValueBuffer("permiso", "rw")
        cursor_flacos.setValueBuffer("idac", id_acs_4)
        cursor_flacos.commitBuffer()

    def test_form_flacos(self) -> None:
        """Test form acls from flacos."""
        from pineboolib.qsa import qsa
        from pineboolib.plugins.mainform import eneboo

        application.ENABLE_ACLS = True

        main_form_class = getattr(eneboo, "MainForm", None)
        self.assertTrue(main_form_class)
        application.PROJECT.main_window = main_form_class()  # type: ignore[misc]

        sys_type = systype.SysType()
        sys_type.installACL("tercera")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)
        button_1 = qsa.from_project("formRecordflmodules").child("botonExportar")  # r-
        button_2 = qsa.from_project("formRecordflmodules").child("botonCargar")  # --
        self.assertTrue(button_2.isHidden())  # not visible
        self.assertFalse(button_2.isEnabled())  # not enabled
        self.assertFalse(button_1.isHidden())  # visible
        self.assertFalse(button_1.isEnabled())  # not enabled

    def test_mainwindow_flacos(self) -> None:
        """Test mainwindow flacos."""
        from PyQt6 import QtGui  # type: ignore[import]
        from pineboolib.plugins.mainform.eneboo import eneboo
        from pineboolib import application

        sys_type = systype.SysType()
        sys_type.installACL("cuarta")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)

        settings.CONFIG.set_value("application/dbadmin_enabled", True)
        application.PROJECT.db_admin_mode = True
        project = application.PROJECT
        # project.main_form = eneboo
        main_form_class = getattr(eneboo, "MainForm", None)
        # main_form_ = getattr(project.main_form, "MainForm", None)
        self.assertTrue(main_form_class)
        self.main_w = main_form_class()  # type: ignore[misc]
        project.main_window = self.main_w
        self.main_w.initScript()
        self.main_w.show()
        self.assertTrue(self.main_w)

        action_1 = self.main_w.findChild(QtGui.QAction, "ebcomportamiento")
        self.assertTrue(action_1)
        self.assertFalse(action_1.isVisible())

        action_2 = self.main_w.findChild(QtGui.QAction, "flgroups")
        self.assertTrue(action_2)
        self.assertFalse(action_2.isVisible())

        action_3 = self.main_w.findChild(QtGui.QAction, "flareas")
        self.assertTrue(action_3)
        self.assertTrue(action_3.isVisible())

    def test_mainwindow_global(self) -> None:
        """Test mainwindow global."""

        from PyQt6 import QtGui
        from pineboolib.plugins.mainform.eneboo import eneboo
        from pineboolib import application

        settings.CONFIG.set_value("ebcomportamiento/main_form_name", "eneboo")

        sys_type = systype.SysType()
        sys_type.installACL("cuarta")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)

        project = application.PROJECT
        main_form_class = getattr(eneboo, "MainForm", None)
        # main_form_ = getattr(project.main_form, "MainForm", None)
        self.assertTrue(main_form_class)
        self.main_w = main_form_class()  # type: ignore[misc]
        project.main_window = self.main_w
        self.main_w.initScript()
        self.main_w.show()
        self.assertTrue(self.main_w)

        action_1 = self.main_w.findChild(QtGui.QAction, "flusers")
        self.assertTrue(action_1)
        self.assertFalse(action_1.isVisible())

    def test_form_globals(self) -> None:
        """Test form acls globals."""
        from pineboolib.qsa import qsa
        from pineboolib import application
        from pineboolib.plugins.mainform.eneboo import eneboo

        sys_type = systype.SysType()
        sys_type.installACL("tercera")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)

        project = application.PROJECT
        main_form_class = getattr(eneboo, "MainForm", None)
        # main_form_ = getattr(project.main_form, "MainForm", None)
        self.assertTrue(main_form_class)
        self.main_w = main_form_class()  # type: ignore[misc]
        project.main_window = self.main_w
        self.main_w.initScript()
        self.main_w.show()
        self.assertTrue(self.main_w)

        form = qsa.from_project("formflmodules")
        control_1 = form.child("tableDBRecords")
        self.assertTrue(control_1)
        self.assertFalse(control_1.isEnabled())

    def test_tables_flacos(self) -> None:
        """Test table acls from flacos."""

        sys_type = systype.SysType()
        sys_type.installACL("primera")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)

        mtd_flgroups = application.PROJECT.conn_manager.manager().metadata("flgroups")

        self.assertTrue(mtd_flgroups is not None)
        # descripcion = '--'
        field_descripcion = mtd_flgroups.field(  # type: ignore [union-attr] # noqa: F821
            "descripcion"
        )
        self.assertFalse(field_descripcion.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field_descripcion.visible())  # type: ignore [union-attr] # noqa: F821

        # idgroup = 'r-'
        field_idgroup = mtd_flgroups.field("idgroup")  # type: ignore [union-attr] # noqa: F821
        self.assertFalse(field_idgroup.visible())  # type: ignore [union-attr] # noqa: F821

        mtd_flmodules = application.PROJECT.conn_manager.manager().metadata("flmodules")
        field_descripcion = mtd_flmodules.field(  # type: ignore [union-attr] # noqa: F821
            "descripcion"
        )

        # descripcion = 'rw'

        self.assertTrue(field_descripcion.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field_descripcion.visible())  # type: ignore [union-attr] # noqa: F821

    def test_tables_globals(self) -> None:
        """Test table acls globals."""

        sys_type = systype.SysType()
        sys_type.installACL("segunda")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)

        mtd_flareas = application.PROJECT.conn_manager.manager().metadata("flareas")
        self.assertTrue(mtd_flareas is not None)
        # '--'
        field_descripcion = mtd_flareas.field(  # type: ignore [union-attr] # noqa: F821
            "descripcion"
        )
        self.assertFalse(field_descripcion.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertFalse(field_descripcion.visible())  # type: ignore [union-attr] # noqa: F821

        mtd_flusers = application.PROJECT.conn_manager.manager().metadata("flusers")
        self.assertTrue(mtd_flusers)
        # 'r-'
        field_descripcion = mtd_flusers.field(  # type: ignore [union-attr] # noqa: F821
            "descripcion"
        )
        self.assertFalse(field_descripcion.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field_descripcion.visible())  # type: ignore [union-attr] # noqa: F821

        mtd_fltest = application.PROJECT.conn_manager.manager().metadata("fltest")
        self.assertTrue(mtd_fltest)
        # 'rw'
        field = mtd_fltest.field("date_field")  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field.visible())  # type: ignore [union-attr] # noqa: F821

    def test_disable_acls(self) -> None:
        """Check if acls.xml load is disables"""

        application.ENABLE_ACLS = False
        sys_type = systype.SysType()
        sys_type.installACL("segunda")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.aq_app.set_acl(acl)

        mtd_flareas = application.PROJECT.conn_manager.manager().metadata("flareas")
        self.assertTrue(mtd_flareas is not None)
        # '--'
        field_descripcion = mtd_flareas.field(  # type: ignore [union-attr] # noqa: F821
            "descripcion"
        )
        self.assertTrue(field_descripcion.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field_descripcion.visible())  # type: ignore [union-attr] # noqa: F821

        mtd_flusers = application.PROJECT.conn_manager.manager().metadata("flusers")
        self.assertTrue(mtd_flusers)
        # 'r-'
        field_descripcion = mtd_flusers.field(  # type: ignore [union-attr] # noqa: F821
            "descripcion"
        )
        self.assertTrue(field_descripcion.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field_descripcion.visible())  # type: ignore [union-attr] # noqa: F821

        mtd_fltest = application.PROJECT.conn_manager.manager().metadata("fltest")
        self.assertTrue(mtd_fltest)
        # 'rw'
        field = mtd_fltest.field("date_field")  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field.editable())  # type: ignore [union-attr] # noqa: F821
        self.assertTrue(field.visible())  # type: ignore [union-attr] # noqa: F821

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()

        """
        from pineboolib.application.database import utils

        sys_type = systype.SysType()
        sys_type.installACL("final")
        acl = pnaccesscontrollists.PNAccessControlLists()
        acl.init()
        application.PROJECT.conn_manager.set_acl(acl)

        utils.sqlDelete("flacls", "1=1")
        utils.sqlDelete("flacs", "1=1")
        utils.sqlDelete("flacos", "1=1")
        utils.sqlDelete("flusers", "1=1")
        utils.sqlDelete("flgroups", "1=1")
        utils.sqlDelete("flfiles", "nombre='acl.xml'")
        application.PROJECT.conn_manager.manager().cacheMetaDataSys_ = {}
        application.PROJECT.conn_manager.manager().cacheMetaData_ = {}
        application.PROJECT.aq_app.acl_ = None

        settings.CONFIG.set_value("application/dbadmin_enabled", cls.db_admin)
        """



"""System Module."""

import os


class System(object):
    """
    System methods.

    For now only envirnoment vars.
    """

    @staticmethod
    def setenv(name: str, val: str) -> None:
        """Set environment variable."""
        os.environ[name] = val

    @staticmethod
    def getenv(name: str) -> str:
        """Get environment variable."""

        return os.environ[name] if name in os.environ.keys() else ""



"""
Import dgi_schema to interface package.
"""
from pineboolib.plugins.dgi import dgi_schema as dgi_schema_module  # pragma: no cover

dgi_schema = dgi_schema_module.DgiSchema  # pragma: no cover # pylint: disable=invalid-name



"""Messagebox module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]

from pineboolib import application
from pineboolib.core.utils import utils_base
from pineboolib.core.utils import logging

from typing import Optional, List

LOGGER = logging.get_logger(__name__)


class MessageBox:
    """MessageBox class."""

    Yes = QtWidgets.QMessageBox.StandardButton.Yes
    No = QtWidgets.QMessageBox.StandardButton.No
    NoButton = QtWidgets.QMessageBox.StandardButton.NoButton
    Ok = QtWidgets.QMessageBox.StandardButton.Ok
    Cancel = QtWidgets.QMessageBox.StandardButton.Cancel
    Ignore = QtWidgets.QMessageBox.StandardButton.Ignore

    @classmethod
    def msgbox(cls, typename, *args) -> Optional["QtWidgets.QMessageBox.StandardButton"]:
        """Return a messageBox."""

        if not getattr(application, "TESTING_MODE", None):
            if utils_base.is_library():
                LOGGER.warning(
                    "q3widget.MessageBox launch when library mode ON! (%s : %s)",
                    typename,
                    args,
                    stack_info=True,
                )
                if typename in ["warning", "critical"]:
                    raise Exception(str(args[0]))
                return None

        msg_box = getattr(QtWidgets.QMessageBox, typename, None)
        title = "Pineboo"
        parent = QtWidgets.QApplication.activeWindow()
        buttons: List["QtWidgets.QMessageBox.StandardButton"] = []
        default_button = None
        text = ""

        for number, argument in enumerate(args):
            if number == 0:
                text = argument
            else:
                if isinstance(argument, str):
                    title = argument
                elif isinstance(argument, QtWidgets.QMessageBox.StandardButton):
                    if len(buttons) < 2:
                        buttons.append(argument)
                    else:
                        default_button = argument
                elif argument:
                    parent = argument

        if application.PROJECT._splash:
            application.PROJECT._splash.hide()

        if msg_box and not getattr(application, "TESTING_MODE", None):
            if not default_button:
                return msg_box(parent, title, text, *buttons)
            else:
                return msg_box(parent, title, text, *buttons, default_button)
        else:
            return QtWidgets.QMessageBox.StandardButton.Ok

    @classmethod
    def question(cls, *args) -> Optional["QtWidgets.QMessageBox.StandardButton"]:
        """Return an question messageBox."""

        return cls.msgbox("question", *args)

    @classmethod
    def information(cls, *args) -> Optional["QtWidgets.QMessageBox.StandardButton"]:
        """Return an information messageBox."""
        return cls.msgbox("information", *args)

    @classmethod
    def warning(cls, *args) -> Optional["QtWidgets.QMessageBox.StandardButton"]:
        """Return an warning messageBox."""

        text_ = args[0] if isinstance(args[0], str) else args[2]

        result = cls.msgbox("warning", *args)
        if not getattr(application, "TESTING_MODE", None):
            clip_board = QtWidgets.QApplication.clipboard()
            if clip_board:
                clip_board.clear()
                clip_board.setText(str(text_))
        return result

    @classmethod
    def critical(cls, *args) -> Optional["QtWidgets.QMessageBox.StandardButton"]:
        """Return an critical messageBox."""

        text_ = args[0] if isinstance(args[0], str) else args[2]
        result = cls.msgbox("critical", *args)
        if not getattr(application, "TESTING_MODE", None):
            clip_board = QtWidgets.QApplication.clipboard()
            if clip_board:
                clip_board.clear()
                clip_board.setText(str(text_))
        return result



"""Flinvalidator module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtGui  # type: ignore
from typing import Any, Tuple, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # pragma: no cover


class FLIntValidator(QtGui.QIntValidator):
    """FLIntValidator Class."""

    _formatting: bool

    def __init__(self, minimum: int, maximum: int, parent: "QtWidgets.QWidget") -> None:
        """Inicialize."""

        super().__init__(minimum, maximum, parent)
        self._formatting = False

    def validate(self, input_: Optional[str], pos_cursor: int) -> Tuple[Any, str, int]:
        """Return validate result."""

        if not input_ or self._formatting:
            return (self.State.Acceptable, input_, pos_cursor)  # type: ignore [return-value]

        state = super().validate(input_, pos_cursor)

        ret_0 = None
        ret_1 = state[1]
        ret_2 = state[2]

        if state[0] in (self.State.Invalid, self.State.Intermediate) and len(input_) > 0:
            text_ = input_[1:]
            if (
                input_[0] == "-"
                and super().validate(text_, pos_cursor)[0] == self.State.Acceptable
                or not text_
            ):
                ret_0 = self.State.Acceptable
            else:
                ret_0 = self.State.Invalid
        else:
            ret_0 = self.State.Acceptable

        return (ret_0, ret_1, ret_2)



"""Dgi_fcgi module."""
# # -*- coding: utf-8 -*-
from pineboolib import logging
from pineboolib.plugins.dgi import dgi_schema
from pineboolib.application.utils.check_dependencies import check_dependencies
from pineboolib.fllegacy import systype

from pineboolib import application

from typing import Any, Mapping


LOGGER = logging.get_logger(__name__)


class DgiFcgi(dgi_schema.DgiSchema):
    """Dgi_fcgi class."""

    _fcgiCall: str
    _fcgiSocket: str

    def __init__(self) -> None:
        """Inicialize."""
        super().__init__()  # desktopEnabled y mlDefault a True
        self._name = "fcgi"
        self._alias = "FastCGI"
        self._fcgi_call = "flfactppal.iface.fcgiProcessRequest"
        self._fcgi_socket = "pineboo-fastcgi.socket"
        self.setUseDesktop(False)
        self.setUseMLDefault(False)
        self.showInitBanner()
        check_dependencies({"flup": "flup-py3"})

    def alternativeMain(self, main_) -> Any:
        """Process alternative main."""
        from flup.server.fcgi import WSGIServer  # type: ignore

        LOGGER.info("=============================================")
        LOGGER.info("FCGI:INFO: Listening socket %s", self._fcgi_socket)
        LOGGER.info("FCGI:INFO: Sending queries to %s", self._fcgi_call)
        par_ = Parser(main_, self._fcgi_call)
        WSGIServer(par_.call, bindAddress=self._fcgi_socket).run()

    def setParameter(self, param: str) -> None:
        """Set parameters."""
        if param.find(":") > -1:
            params_list = param.split(":")
            self._fcgi_call = params_list[0]
            self._fcgi_socket = params_list[1]
        else:
            self._fcgi_call = param


"""
Esta clase lanza contra el arbol qsa la consulta recibida y retorna la respuesta proporcionada, si procede
"""


class Parser(object):
    """Parser class."""

    _prj = None
    _call_script: str

    def __init__(self, prj: Any, callScript: str) -> None:
        """Inicialize."""
        self._prj = prj
        self._call_script = callScript

    def call(self, environ: Mapping[str, Any], start_response) -> Any:
        """Return value from called function."""
        start_response("200 OK", [("Content-Type", "text/html")])
        query_str = environ["QUERY_STRING"]
        try:
            retorno_: Any = application.PROJECT.call(self._call_script, query_str)
        except Exception:
            qsa_sys = systype.SysType()

            LOGGER.info(self._call_script, environ["QUERY_STRING"])
            retorno_ = (
                """<html><head><title>Pineboo %s - FastCGI - </title></head><body><h1>Function %s not found!</h1></body></html>"""
                % (qsa_sys.version(), self._call_script)
            )
            pass
        LOGGER.info("FCGI:INFO: Processing '%s' ...", environ["QUERY_STRING"])

        return retorno_



"""Test_pnconnection module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.application.database import pnsqlcursor


class TestPNConnection(unittest.TestCase):
    """TestPNConnection Class."""

    @classmethod
    def setUp(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic1(self) -> None:
        """Basic test 1."""

        conn_manager = application.PROJECT.conn_manager

        self.assertEqual(conn_manager.mainConn().connectionName(), "main_conn")
        conn_default_ = conn_manager.useConn("default")
        conn_db_aux_ = conn_manager.useConn("dbAux")
        conn_aux = conn_manager.useConn("Aux")
        conn_ = conn_manager.useConn("conn_test")
        self.assertNotEqual(conn_, conn_db_aux_)
        self.assertNotEqual(conn_db_aux_, conn_aux)
        dict_databases_1 = conn_manager.enumerate()
        self.assertTrue(dict_databases_1)

        self.assertTrue(conn_default_.isOpen())
        self.assertTrue(conn_manager.mainConn().isOpen())
        self.assertTrue(conn_manager.useConn("dbAux").isOpen())
        self.assertTrue(conn_manager.useConn("Aux").isOpen())
        self.assertFalse(conn_manager.useConn("conn_test").isOpen())

        self.assertEqual([*dict_databases_1], ["default", "dbAux", "Aux", "conn_test"])
        self.assertTrue("flareas" in conn_aux.tables("Tables"))
        self.assertTrue(conn_manager.removeConn("conn_test"))
        dict_databases_2 = conn_manager.enumerate()
        self.assertEqual([*dict_databases_2], ["default", "dbAux", "Aux"])

    def test_basic2(self) -> None:
        """Basic test 2."""

        conn_manager = application.PROJECT.conn_manager
        conn_ = conn_manager.default()
        self.assertTrue("flareas" in conn_.tables("Tables"))
        self.assertTrue("sqlite_master" in conn_.tables())
        self.assertEqual(conn_.tables("SystemTables"), ["sqlite_master"])
        self.assertEqual(conn_.tables("Views"), [])

        data_base = conn_.database()
        data_base_aux = conn_manager.database("dbAux")
        self.assertNotEqual(
            conn_manager.db(), conn_manager.mainConn()
        )  # Compares default Vs main_conn
        self.assertNotEqual(data_base, data_base_aux)
        self.assertEqual(conn_.DBName(), str(conn_))

    def test_basic3(self) -> None:
        """Basic test 3."""

        from pineboolib.fllegacy import systype

        conn_manager = application.PROJECT.conn_manager
        sys_type = systype.SysType()
        conn_ = conn_manager.mainConn()

        self.assertTrue(conn_.driver())
        self.assertTrue(conn_.driver().connection())
        sys_type.addDatabase("conn_test_2")
        self.assertTrue(conn_manager.useConn("conn_test_2").isOpen())
        self.assertTrue(sys_type.removeDatabase("conn_test_2"))
        self.assertEqual(conn_.driverName(), "FLsqlite")
        self.assertEqual(conn_.driver().alias_, conn_.driverAlias())
        self.assertEqual(conn_.driverNameToDriverAlias(conn_.driverName()), "SQLite3 (SQLITE3)")

    def test_basic4(self) -> None:
        """Basic test 4."""

        conn_manager = application.PROJECT.conn_manager
        conn_ = conn_manager.default()
        self.assertTrue(conn_.interactiveGUI())
        conn_.setInteractiveGUI(False)
        self.assertFalse(conn_.interactiveGUI())
        # self.assertNotEqual(conn_, conn_manager.db())
        self.assertEqual(conn_manager.dbAux(), conn_manager.useConn("dbAux"))

        self.assertEqual(conn_.formatValue("string", "hola", True), "'HOLA'")
        self.assertEqual(conn_.formatValueLike("string", "hola", True), "LIKE 'HOLA%%'")
        # self.assertTrue(conn_.canSavePoint())
        # self.assertTrue(conn_.canTransaction())
        self.assertEqual(conn_.transactionLevel(), 0)
        self.assertTrue(conn_.canDetectLocks())
        self.assertTrue(conn_manager.manager())
        self.assertTrue(conn_manager.managerModules())
        self.assertTrue(conn_.canOverPartition())

        mtd_seqs = conn_manager.manager().metadata("flseqs")
        self.assertTrue(mtd_seqs is not None)
        if mtd_seqs is not None:
            self.assertFalse(conn_.existsTable("fltest"))
            self.assertTrue(conn_.existsTable("flseqs"))
            self.assertFalse(conn_.mismatchedTable("flseqs", mtd_seqs))
            self.assertEqual(conn_.normalizeValue("holá, 'avión'"), "holá, ''avión''")

    def test_basic5(self) -> None:
        """Basic test 5."""

        conn_manager = application.PROJECT.conn_manager
        conn_ = conn_manager.useConn("default")
        cursor = pnsqlcursor.PNSqlCursor("flareas")
        conn_.doTransaction(cursor)
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("idarea", "test")
        cursor.setValueBuffer("bloqueo", "false")
        cursor.setValueBuffer("descripcion", "test area")
        self.assertTrue(cursor.commitBuffer())
        conn_.doRollback(cursor)
        conn_.doTransaction(cursor)
        cursor.setModeAccess(cursor.Insert)
        cursor.setValueBuffer("idarea", "test")
        cursor.setValueBuffer("bloqueo", "false")
        cursor.setValueBuffer("descripcion", "test area")
        cursor.commitBuffer()
        conn_.doCommit(cursor, False)
        conn_.canRegenTables()

        self.assertEqual(conn_.tables(1)[0:3], ["flareas", "flfiles", "flgroups"])

        self.assertEqual(conn_.tables(2), ["sqlite_master"])
        self.assertEqual(conn_.tables(3), [])

        self.assertTrue(conn_.session())
        self.assertTrue(conn_.engine())

        self.assertFalse(conn_.port())
        self.assertFalse(conn_.returnword())

        # self.assertFalse(conn_.lastActiveCursor())

        # conn_.Mr_Proper() #FIXME

    def test_basic6(self) -> None:
        """Test basic 6."""
        from pineboolib.application.database import pnsqlcursor

        conn_manager = application.PROJECT.conn_manager
        conn_default = conn_manager.useConn("default")
        conn_manager.useConn("test")
        cursor = pnsqlcursor.PNSqlCursor("flsettings")
        cursor.setAskForCancelChanges(False)
        conn_manager.mainConn().Mr_Proper()
        # self.assertEqual(
        #    conn_manager.mainConn().queryUpdate("test", "field1, 'val_1'", "1=1"),
        #    "UPDATE test SET field1, 'val_1' WHERE 1=1",
        # )
        self.assertTrue(conn_manager.removeConn("test"))
        self.assertTrue(conn_default.doTransaction(cursor))
        self.assertTrue(cursor.inTransaction())
        self.assertTrue(conn_default.doCommit(cursor, False))
        self.assertTrue(conn_default.doTransaction(cursor))
        self.assertTrue(conn_default.doRollback(cursor))

    def test_reinit_connections(self) -> None:
        """Test removing users connections."""

        application.PROJECT.conn_manager.reinit_user_connections()
        self.assertTrue(application.PROJECT.conn_manager.default().isOpen())

    @classmethod
    def tearDown(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestPNConnectionIsolation(unittest.TestCase):
    """TestPNConnectionIsolation Class."""

    @classmethod
    def setUp(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        # application.LOG_SQL = True
        init_testing()

    def test_isolation_sessions(self) -> None:
        """Test isolated sessions."""

        conn_manager = application.PROJECT.conn_manager
        default = conn_manager.default()
        db_aux = conn_manager.dbAux()
        self.assertNotEqual(default.session(), db_aux.session())

    def test_transactions(self) -> None:
        """Test transactions effects."""

        cursor_default = pnsqlcursor.PNSqlCursor("fltest5", "default")
        cursor_default.transaction()
        self.assertEqual(cursor_default.transactionLevel(), 1)
        cursor_dbaux = pnsqlcursor.PNSqlCursor("fltest5", "dbaux")
        self.assertEqual(cursor_dbaux.transactionLevel(), 0)
        cursor_dbaux.transaction()
        self.assertEqual(cursor_dbaux.transactionLevel(), 1)
        cursor_dbaux.commit()
        self.assertEqual(cursor_dbaux.transactionLevel(), 0)
        self.assertEqual(cursor_default.transactionLevel(), 1)
        cursor_default.transaction()
        self.assertEqual(cursor_default.transactionLevel(), 2)
        cursor_default.commit()
        self.assertEqual(cursor_default.transactionLevel(), 1)
        cursor_default.commit()
        self.assertEqual(cursor_default.transactionLevel(), 0)

    def test_out_transactions(self) -> None:
        """Test out_transactions effects."""

        cursor_aux = pnsqlcursor.PNSqlCursor("fltest5", "dbAux")
        cursor_aux.transaction()  # aux 1
        cursor_aux.setModeAccess(cursor_aux.Insert)
        cursor_aux.refreshBuffer()

        cursor_aux.transaction()  # aux 2
        default_conn = application.PROJECT.conn_manager.default()
        default_conn.transaction()  # default 1

        self.assertEqual(cursor_aux.transactionLevel(), 2)

        cursor = pnsqlcursor.PNSqlCursor("fltest3")
        cursor.transaction()  # default 2
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.select()
        self.assertEqual(cursor.size(), 1)
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Edit)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bool_field", True)

        self.assertEqual(cursor_aux.transactionLevel(), 2)

        cursor_aux.commit()  # aux 2

        self.assertTrue(cursor.commitBuffer())
        cursor_aux.commit()  # aux 1
        self.assertEqual(cursor_aux.transactionLevel(), 0)

    @classmethod
    def tearDown(cls) -> None:
        """Ensure test clear all data."""
        # application.LOG_SQL = False
        finish_testing()



"""Test_pncursortablemodel module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.database import pnsqlcursor


class TestPNCursorTableModel(unittest.TestCase):
    """TestPNCursorTableModel Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Basic test 1."""

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("string_field", "xxx")
        cursor.setValueBuffer("double_field", 0.02)
        cursor.setValueBuffer("date_field", "2019-01-01")
        cursor.commitBuffer()
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("string_field", "zzz")
        cursor.setValueBuffer("double_field", 0.01)
        cursor.commitBuffer()
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("string_field", "yyy")
        cursor.setValueBuffer("double_field", 0.03)

        # cursor.setValueBuffer("check_field", True)
        cursor.commitBuffer()
        cursor.select()
        cursor.setSort("string_field ASC")
        cursor.last()
        self.assertEqual(cursor.currentRegister(), 2)
        cursor.select()
        self.assertTrue(cursor.last())
        self.assertEqual(cursor.valueBuffer("string_field"), "zzz")
        self.assertEqual(cursor.valueBuffer("double_field"), 0.01)
        cursor.prev()
        self.assertEqual(cursor.valueBuffer("string_field"), "yyy")

    def test_basic_2(self) -> None:
        """Basic test 2."""

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.select()
        cursor.last()
        cursor.refreshBuffer()
        self.assertEqual(cursor.valueBuffer("string_field"), "yyy")

        model = cursor.model()

        self.assertEqual(model.find_pk_row(cursor.valueBuffer("id")), cursor.size() - 1)
        self.assertEqual(model.metadata().primaryKey(), "id")
        self.assertEqual(model.fieldType("string_field"), "string")
        self.assertEqual(model.alias("string_field"), "String field")
        self.assertEqual(
            model.metadata().field("string_field"), cursor.metadata().field("string_field")
        )

    def test_basic_3(self) -> None:
        from PyQt6 import QtCore  # type: ignore[import]
        import locale
        import os
        from datetime import date

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setSort("string_field DESC")

        cursor.select()
        model = cursor.model()

        self.assertEqual(model.data(model.index(0, 1)), "zzz")
        self.assertEqual(model.data(model.index(0, 0)), 4)
        self.assertEqual(model.data(model.index(0, 2)), None)
        self.assertEqual(
            model.data(model.index(0, 4)), QtCore.QLocale.system().toString(float(0.01), "f", 2)
        )
        self.assertEqual(model.data(model.index(0, 5)), "No")
        self.assertEqual(model.data(model.index(1, 1)), "yyy")
        self.assertEqual(model.data(model.index(1, 0)), 6)

        cursor.setSort("string_field DESC, double_field DESC")
        model.sort(0, QtCore.Qt.SortOrder.AscendingOrder)
        self.assertEqual(
            model.data(model.index(0, 5), QtCore.Qt.ItemDataRole.TextAlignmentRole),
            QtCore.Qt.AlignmentFlag.AlignVCenter | QtCore.Qt.AlignmentFlag.AlignCenter,
        )
        self.assertEqual(
            model.data(model.index(1, 1), QtCore.Qt.ItemDataRole.TextAlignmentRole),
            QtCore.Qt.AlignmentFlag.AlignVCenter,
        )
        system_date = date(2019, 1, 1)
        locale.setlocale(locale.LC_TIME, "")
        date_format = "%%d/%%m/%%y" if os.name == "nt" else locale.nl_langinfo(locale.D_FMT)

        date_format = date_format.replace("y", "Y")  # Año con 4 dígitos
        date_format = date_format.replace("/", "-")  # Separadores
        date_ = system_date.strftime(date_format)

        self.assertEqual(model.data(model.index(0, 2), QtCore.Qt.ItemDataRole.DisplayRole), date_)
        self.assertEqual(
            model.data(model.index(1, 4), QtCore.Qt.ItemDataRole.DisplayRole),
            QtCore.QLocale.system().toString(float(0.01), "f", 2),
        )

    def test_basic_4(self) -> None:
        """Test basic 4."""
        from PyQt6 import QtCore, QtGui

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setSort("string_field DESC")
        cursor.select()

        model = cursor.model()
        self.assertTrue(
            isinstance(
                model.data(model.index(0, 5), QtCore.Qt.ItemDataRole.BackgroundRole), QtGui.QBrush
            )
        )
        self.assertTrue(
            isinstance(
                model.data(model.index(0, 5), QtCore.Qt.ItemDataRole.ForegroundRole), QtGui.QBrush
            )
        )

        model.update_rows()
        # self.assertFalse(model.findCKRow([]))
        # self.assertFalse(model.findCKRow([2, 2]))
        self.assertEqual(model.find_pk_row(21), -1)
        self.assertEqual(model.find_pk_row(4), 0)

    def test_basic_5(self) -> None:
        """Test basic 5."""
        from PyQt6 import QtCore

        cursor = pnsqlcursor.PNSqlCursor("fltest2")

        model = cursor.model()
        model.disable_refresh(True)
        model.sort(1, QtCore.Qt.SortOrder.DescendingOrder)
        self.assertTrue(model._disable_refresh)
        model.disable_refresh(False)
        model.update_rows()
        cursor.select()
        model.updateColumnsCount()
        self.assertEqual(model.rowCount(), 3)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestFetchMore(unittest.TestCase):
    """Test Acos class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Test fetchmoderows 1."""

        from pineboolib.application.database import pnsqlcursor

        cur_test = pnsqlcursor.PNSqlCursor("fltest")
        size = 2102
        for i in range(size):
            cur_test.setModeAccess(cur_test.Insert)
            self.assertTrue(cur_test.refreshBuffer())
            cur_test.setValueBuffer("string_field", "Registro %s" % i)
            self.assertTrue(cur_test.commitBuffer())

        cur_test.select()
        self.assertEqual(cur_test.size(), size)

    def test_basic_2(self) -> None:
        """Test fetchmoderows 2."""

        from pineboolib.application.database import pnsqlcursor

        cur_test = pnsqlcursor.PNSqlCursor("fltest")
        cur_test.select()
        self.assertTrue(cur_test.first())
        self.assertEqual(cur_test.valueBuffer("string_field"), "Registro 0")
        self.assertTrue(cur_test.next())
        self.assertEqual(cur_test.valueBuffer("string_field"), "Registro 1")
        self.assertTrue(cur_test.last())
        self.assertEqual(cur_test.valueBuffer("string_field"), "Registro 2101")
        self.assertTrue(cur_test.prev())
        self.assertEqual(cur_test.valueBuffer("string_field"), "Registro 2100")

    def test_basic_3(self) -> None:
        """Test fetchmoderows 3."""

        from pineboolib.application.database import pnsqlquery

        qry_test = pnsqlquery.PNSqlQuery()
        qry_test.setTablesList("fltest")
        qry_test.setFrom("fltest")
        qry_test.setWhere("1=1")
        qry_test.setSelect("string_field")
        qry_test.setOrderBy("id")
        self.assertTrue(qry_test.exec_())
        self.assertEqual(qry_test.size(), 2102)
        self.assertTrue(qry_test.first())
        self.assertEqual(qry_test.value(0), "Registro 0")
        self.assertTrue(qry_test.last())
        self.assertEqual(qry_test.value(0), "Registro 2101")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Basemodel module."""

from pineboolib.core import decorators
from pineboolib.core.utils import logging
from pineboolib.application.metadata import pnrelationmetadata
from pineboolib.application import qsadictmodules
from pineboolib.application.database.orm.utils import do_flush
from pineboolib import application

from pineboolib.application.database.orm import dummy_cursor, dummy_signal

from typing import Optional, List, Dict, Union, Callable, Any, TYPE_CHECKING

from sqlalchemy import orm, inspect  # type: ignore [import]
import datetime
import sys
import time
import types

if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata  # noqa: F401 # pragma: no cover
    from pineboolib.application import xmlaction  # noqa: F401 # pragma: no cover


LOGGER = logging.get_logger(__name__)

relation_proxy_objects: Dict[str, List] = {}


class Copy:
    """Copy class."""

    pass


class BaseModel(object):
    """Base Model class."""

    __tablename__: str = ""
    __mapper_args__: Dict[str, Any] = {"confirm_deleted_rows": False}

    _session: Optional["orm.session.Session"]
    _buffer_copy: "Copy"
    _result_before_flush: bool
    _result_after_flush: bool
    _force_mode: Optional[int]
    _current_mode: Optional[int]
    _cursor: "dummy_cursor.DummyCursor"
    _before_commit_function: str
    _after_commit_function: str
    _new_object: bool
    _deny_buffer_changed: List[str]
    bufferChanged: "dummy_signal.FakeSignal"
    _action: Optional["xmlaction.XMLAction"]
    legacy_metadata: Dict[str, Any]
    _cached_bufferchanged: Dict[str, Any]
    serial: bool = True
    counter: bool = False
    no_init: bool = False

    @classmethod
    def _constructor_init(cls, target, kwargs={}) -> None:
        cls._from_query_init(target)

    @classmethod
    def _from_query_init(cls, target):
        target._session = inspect(target).session
        target._action = None

        if not target._session:
            cls._error_manager("_constructor_init", "session is empty!")

        elif not target._session._conn_name:
            cls._error_manager("_constructor_init", "session is invalid!")

        target._new_object = False
        target._common_init()

    def _qsa_init(target, args=[], kwargs={}) -> None:
        """Initialize from qsa."""

        target._session = None
        target._action = None
        conn_name = "default"
        if "session" in kwargs:
            target._session = kwargs["session"]
        else:
            conn_manager = application.PROJECT.conn_manager
            if "conn_name" in kwargs.keys():
                conn_name = kwargs["conn_name"]
            target._session = conn_manager.useConn(conn_name).session()
        if target._session is None:
            target._error_manager(
                "_qsa_init",
                "An active thread session or atomic session was not found on the '%s' connection."
                % (conn_name),
            )

        target._new_object = True
        target.counter = False
        target.no_init = False
        target.serial = True

        if "serial" in kwargs:
            target.serial = kwargs["serial"]

        if "counter" in kwargs:
            target.counter = kwargs["counter"]

        if "no_init" in kwargs:
            target.no_init = kwargs["no_init"]

        for key, value in kwargs.items():
            if hasattr(target, key):
                setattr(target, key, value)

        target._common_init()

    @classmethod
    def get_session_from_connection(cls, conn_name: str = "default") -> "orm.session.Session":
        """Return new session from a connection."""
        new_session = application.PROJECT.conn_manager.useConn(conn_name).session()
        # setattr(new_session, "_conn_name", conn_name)
        return new_session

    def load_action(self) -> None:
        """Load action."""

        self._action = application.PROJECT.actions[self.__tablename__]
        if self._action is not None:
            if self._action._record_script and not self._action._record_widget:
                self._action.load_record_widget()

    def _common_init(self) -> None:
        """Initialize."""
        self.bufferChanged = dummy_signal.FakeSignal(self)  # pylint: disable=invalid-name
        self._force_mode = None
        self._cached_bufferchanged = {}

        if self.__tablename__ in application.PROJECT.actions.keys():
            self.load_action()

        # else:
        #    self._error_manager(
        #        "_common_init",
        #        "%s no se encuentra en %s"
        #        % (self.__tablename__, application.PROJECT.actions.keys()),
        #    )

        self._deny_buffer_changed = []

        if not self._session:
            self._error_manager("_common_init", "session is empty!")
        else:
            if not self._session._conn_name:  # type: ignore [attr-defined] # noqa: F821
                self._error_manager("_common_init", "Session_name not found!")

            if self in self._session.new:
                self._error_manager("_common_init", "Common init with session.new instance!")

            if not hasattr(self, "_buffer_copy"):
                self.update_copy()
                self._current_mode = None
                # self._force_mode =  3 #browse

            if self._new_object:
                self._populate_default()

            self._cursor = dummy_cursor.DummyCursor(self)

            self._before_commit_function = "beforeCommit_%s" % self.__tablename__
            self._after_commit_function = "afterCommit_%s" % self.__tablename__

            try:
                if self._new_object:
                    if self.serial:
                        self.init_serial()

                    if self.counter:
                        for field in self.legacy_metadata["fields"]:
                            if "counter" in field.keys() and field["counter"]:
                                self.init_counter(field["name"])

                    if self._action and self._action._record_widget is not None:
                        iface = getattr(
                            self._action._record_widget, "iface", self._action._record_widget
                        )
                        if iface is not None and not self.no_init:
                            func_ = getattr(iface, "iniciaValoresCursor", None)
                            if func_ is not None:
                                try:
                                    func_(self.cursor)
                                except Exception as error:
                                    self._error_manager("_common_init.iniciaValoresCursor", error)

                    self.init_new()
                self.init()
            except Exception as error:
                self._error_manager("_common_init", error)

    def _validate_cursor(self) -> None:
        """Validate cursor."""

        if self._action is not None and self._action._record_widget is not None:
            iface = getattr(self._action._record_widget, "iface", self._action._record_widget)
            if iface is not None:
                func_ = getattr(iface, "validateCursor", None)
                if func_ is not None:
                    result = True
                    try:
                        result = func_(self.cursor)
                    except Exception as error:
                        self._error_manager("_common_init.validateCursor", error)

                    if result is False:
                        self._error_manager(
                            "_common_init.validateCursor", "validateCursor returned False"
                        )

    def init(self):
        """Initialize."""
        # print("--->", self, self._session)
        pass

    def copy(self) -> "Copy":
        """Return buffer_copy."""

        return self._buffer_copy

    def update_copy(self) -> None:
        """Update buffer copy."""
        self._buffer_copy = Copy()

        # table_mtd = self.table_metadata()

        while not self._new_object and self.pk is None:
            time.sleep(10)

        for field in self.legacy_metadata["fields"]:
            field_name = field["name"]
            setattr(self._buffer_copy, field_name, getattr(self, field_name, None))

    def changes(self) -> Dict[str, Any]:
        """Return field names changed and values."""
        changes = {}

        # table_mtd = self.table_metadata()

        for field in self.legacy_metadata["fields"]:
            field_name = field["name"]
            original_value = getattr(self._buffer_copy, field_name, None)
            current_value = getattr(self, field_name)

            if type(original_value) != type(current_value):  # noqa: E721
                changes[field_name] = current_value
            elif original_value != current_value:
                changes[field_name] = current_value

        return changes

    def after_new(self) -> Optional[bool]:
        """After flush new instance."""

        return True

    def after_change(self) -> Optional[bool]:
        """After update a instance."""

        return True

    def after_delete(self) -> Optional[bool]:
        """After delete a instance."""

        return True

    def after_flush(self) -> Optional[bool]:
        """After flush."""

        return True

    def before_new(self) -> Optional[bool]:
        """Before flush new instance."""

        return True

    def before_change(self) -> Optional[bool]:
        """Before update a instance."""

        return True

    def before_delete(self) -> Optional[bool]:
        """Before delete a instance."""

        return True

    def before_flush(self) -> Optional[bool]:
        """Before flush."""

        return True

    def init_new(self) -> None:
        """Init for new instances."""

        pass

    def delete(self) -> bool:
        """Flush instance to current session."""

        if self._session:
            # if not self._session.in_transaction():
            #    self._session.begin()
            # else:
            #    self._session.begin_nested()
            self._force_mode = 2  # delete.

            self._flush()
        else:
            self._error_manager("delete", "_session is empty!")

        return True

    def _delete_cascade(self) -> None:
        """Delete cascade instances if proceed."""

        for field in self.table_metadata().fieldList():
            relation_list = field.relationList()
            for relation in relation_list:
                foreign_table_mtd = application.PROJECT.conn_manager.manager().metadata(
                    relation.foreignTable()
                )
                if foreign_table_mtd is not None:
                    foreign_field_mtd = foreign_table_mtd.field(relation.foreignField())
                    if foreign_field_mtd is not None:
                        relation_m1 = foreign_field_mtd.relationM1()
                        if relation_m1 is not None and relation_m1.deleteCascade():
                            foreign_table_class = qsadictmodules.QSADictModules.orm_(
                                foreign_table_mtd.name()
                            )
                            if foreign_table_class is not None:
                                foreign_field_object = getattr(
                                    foreign_table_class, relation.foreignField()
                                )
                                relation_objects = (
                                    foreign_table_class.query(
                                        self._session._conn_name  # type: ignore [union-attr] # noqa: F821
                                    )
                                    .filter(foreign_field_object == getattr(self, field.name()))
                                    .all()
                                )

                                for obj in relation_objects:
                                    if not obj.delete():
                                        self._error_manager(
                                            "_delete_cascade",
                                            "obj: %s, pk_value: %s can't deleted" % (obj, obj.pk),
                                        )

    def _flush(
        self, relations: List[str] = [], only: List[str] = [], ignore_foreignkey: bool = False
    ) -> None:
        """Flush data."""

        if self._session is None:
            self._error_manager("_flush", "_session is empty")
        else:
            self._current_mode = self.mode_access
            if not only or "before_flush" in only:
                self._before_flush()

            if not only or "check_integrity" in only:
                self._check_integrity(ignore_foreignkey=ignore_foreignkey)

            if not only or "delete_cascade" in only:
                if self._current_mode == 2:  # delete
                    self._delete_cascade()

                    if (
                        self not in self._session.deleted
                    ):  # hay que hacerlo aquí, despues de before_* , porque si nó, cualquier session.flush borraria el padre.
                        self._session.delete(self)

            if not only or "flush" in only:
                for relation in relations:
                    list_objects = getattr(self, relation, [])

                    for list_object in list_objects:
                        list_object._flush(
                            only=["before_flush", "check_integrity", "delete_cascade"],
                            ignore_foreignkey=True,
                        )
                try:
                    flush_objects = [self]
                    for relation in relations:
                        list_objects = getattr(self, relation, [])
                        for list_object in list_objects:
                            flush_objects.append(list_object)

                    do_flush(self._session, flush_objects)

                except Exception as error:
                    self._error_manager("_flush", error)

                for relation in relations:
                    list_objects = getattr(self, relation, [])
                    for list_object in list_objects:
                        list_object._flush(only=["after_flush"])

            if not only or "after_flush" in only:
                self._after_flush()
            # else:
            #    self._current_mode = 3  # edit
        self._current_mode = None

    def _before_flush(self) -> None:
        """Before flush."""

        try:
            mode = self._current_mode
            func_ = getattr(self.module_iface, self._before_commit_function, None)
            if func_ is not None:
                value = func_(self._cursor)
                if value and not isinstance(value, bool) or value is False:
                    self._error_manager("beforeCommit", "%s return False" % func_)
        except Exception as error:
            self._error_manager("_before_flush", error)

        try:
            self.before_flush()

        except Exception as error:
            self._error_manager("before_flush", error)

        if mode == 0:  # insert
            try:
                self.before_new()
                self._validate_cursor()

            except Exception as error:
                self._error_manager("before_new", error)

        elif mode == 1:  # edit
            try:
                self.before_change()
                self._validate_cursor()

            except Exception as error:
                self._error_manager("before_change", error)

        elif mode == 2:  # delete
            try:
                self.before_delete()
            except Exception as error:
                self._error_manager("before_delete", error)

    def _after_flush(self) -> None:
        """After flush."""

        try:
            mode = self._current_mode

            func_ = getattr(self.module_iface, self._after_commit_function, None)
            if func_ is not None:
                value = func_(self._cursor)
                if value and not isinstance(value, bool) or value is False:
                    self._error_manager("afterCommit", "%s return False" % func_)
        except Exception as error:
            self._error_manager("_after_flush", error)

        try:
            self.after_flush()
        except Exception as error:
            self._error_manager("after_flush", error)

        if mode == 0:  # insert
            try:
                self.after_new()
            except Exception as error:
                self._error_manager("after_new", error)

        elif mode == 1:  # edit
            try:
                self.after_change()
            except Exception as error:
                self._error_manager("after_change", error)

        elif mode == 2:  # delete
            try:
                self.after_delete()
            except Exception as error:
                self._error_manager("after_delete", error)

    @classmethod
    def table_metadata(cls) -> "pntablemetadata.PNTableMetaData":
        """Return table metadata."""

        ret_ = application.PROJECT.conn_manager.manager().metadata(cls.__tablename__)

        if ret_ is None:
            cls._error_manager("table_metadata", "%s tablemetadata is empty" % cls.__tablename__)

        return ret_  # type: ignore [return-value] # noqa: F723

    @classmethod
    def type(cls, field_name: str = ""):
        """Return field type."""

        field_mtd = cls.table_metadata().field(field_name)
        if field_mtd is not None:
            return field_mtd.type()

        return None

    @classmethod
    def get(cls, pk_value: str, session: Union[str, "orm.Session"] = "default") -> Any:
        """Return instance selected by pk."""
        # qry = cls.query(session)
        # ret_ = qry.get(pk_value) if qry is not None else None

        session_ = (
            application.PROJECT.conn_manager.useConn(session).session()
            if isinstance(session, str)
            else session
        )

        return session_.get(cls, pk_value) if session_ else None  # type: ignore [attr-defined]

    @classmethod
    @decorators.deprecated
    def query(
        cls, session_or_name: Union[str, "orm.Session"] = "default"
    ) -> Optional["orm.query.Query"]:
        """Return Session query."""

        ret_ = None
        session_: Optional["orm.session.Session"] = None
        mng_ = application.PROJECT.conn_manager
        if session_or_name is not None:
            if isinstance(session_or_name, str):
                session_ = mng_.useConn(session_or_name).session()
            else:
                session_ = session_or_name

            if isinstance(session_, orm.session.Session):
                ret_ = session_.query(cls)

        if ret_ is None:
            LOGGER.warning(  # type: ignore [unreachable]
                "query: Invalid session %s " % session_or_name
            )

        return ret_

    @classmethod
    def _before_compile_update(cls, query, context) -> bool:
        """Before compile Update."""
        for obj in query.all():
            obj._current_mode = 1  # edit
            obj._before_flush()

            obj._after_flush()

        return True

    @classmethod
    def _before_compile_delete(cls, query, context) -> bool:
        """Before compile Delete."""
        for obj in query.all():
            obj._current_mode = 2  # delete
            obj._before_flush()
            obj._delete_cascade()

            obj._after_flush()

        return True

    def _populate_default(self) -> None:
        """Populate with default values."""

        metadata = self.table_metadata()

        for name in metadata.fieldNames():
            field_mtd = metadata.field(name)

            if field_mtd is None:
                LOGGER.warning("%s metadata not found!", name)
                continue

            default_value = field_mtd.defaultValue()
            if default_value is None:
                continue

            if isinstance(default_value, str):
                type_ = field_mtd.type()

                if type_ == "date":
                    default_value = datetime.date.fromisoformat(str(default_value)[:10])
                elif type_ == "timestamp":
                    default_value = datetime.datetime.strptime(
                        str(default_value), "%Y-%m-%d %H:%M:%S"
                    )
                elif type_ == "time":
                    default_value = str(default_value)
                    if default_value.find("T") > -1:
                        default_value = default_value[default_value.find("T") + 1 :]

                    default_value = datetime.datetime.strptime(
                        str(default_value)[:8], "%H:%M:%S"
                    ).time()

            setattr(self, name, default_value)

    def save(self, relations: List[str] = []) -> bool:
        """Flush instance to current session."""

        if not hasattr(self, "_session"):
            self._error_manager(
                "save", "This new instance was not initialized with qsa.orm_(class_name)"
            )

        else:
            if self._session is None:
                self._error_manager("save", "_session is empty!")
            elif self.mode_access == 2:
                self._error_manager(
                    "save", "you are trying to save an instance in the process of deletion!"
                )
            else:
                if self.mode_access == 0:  # insert
                    self._session.add(self)

                self._flush(relations)

            self.update_copy()
            return True
        return False

    def _check_integrity(self, ignore_foreignkey: bool = False) -> bool:
        """Check data integrity."""

        mode = self.mode_access

        table_meta = self.table_metadata()

        if not table_meta.isQuery():
            for field in table_meta.fieldList():
                field_name = field.name()
                if mode < 2:  # 0 insert,1 edit
                    # not Null fields.
                    if not field.allowNull():
                        value = getattr(self, field_name, None)
                        if value is None:
                            self._error_manager(
                                "_check_integrity",
                                "INTEGRITY::Field %s.%s need a value"
                                % (table_meta.name(), field_name),
                                self,
                            )
                        elif field.type() == "date" and not isinstance(value, datetime.date):
                            self._error_manager(
                                "_check_integrity",
                                "INTEGRITY::Type Error %s.%s -> Value must be a datetime.date type, but found %s type"
                                % (table_meta.name(), field_name, type(value)),
                                self,
                            )
                        elif field.type() == "time" and not isinstance(value, datetime.time):
                            self._error_manager(
                                "_check_integrity",
                                "INTEGRITY::Type Error %s.%s -> Value must be a datetime.time type, but found %s type"
                                % (table_meta.name(), field_name, type(value)),
                                self,
                            )

                # para poder comprobar relaciones , tengo que mirar primero que los campos not null esten ok, si no , da error.

                relation_m1 = field.relationM1()
                if relation_m1 is not None:
                    foreign_class_ = None
                    if qsadictmodules.QSADictModules.action_exists(
                        "%s_orm" % relation_m1.foreignTable()
                    ):
                        foreign_class_ = qsadictmodules.QSADictModules.orm_(
                            relation_m1.foreignTable()
                        )

                    if foreign_class_ is not None:
                        foreign_class_.table_metadata()
                        foreign_field_obj = getattr(
                            foreign_class_, relation_m1.foreignField(), None
                        )

                        value = getattr(self, field_name)

                        if isinstance(value, bool):
                            self._error_manager(
                                "_check_integrity",
                                "INTEGRITY::Field relation %s.%s -> %s A boolean has been assigned (%s).%s"
                                % (
                                    table_meta.name(),
                                    field_name,
                                    foreign_field_obj,
                                    value,
                                    " Use None instead of False" if not value else "",
                                ),
                                self,
                            )

                        qry_data = None
                        try:
                            qry_data = (
                                self._session.query(  # type: ignore [union-attr] # noqa: F821
                                    foreign_class_
                                )
                                .filter(foreign_field_obj == value)
                                .first()
                            )
                        except Exception as error:
                            self._error_manager(
                                "_check_integrity",
                                "INTEGRITY::Field relation %s.%s -> %s"
                                % (table_meta.name(), field_name, error),
                                self,
                            )
                        # qry_data = (
                        #    foreign_class_.query(self._session)
                        #    .filter(foreign_field_obj == getattr(self, field_name))
                        #    .first()
                        # )

                        if isinstance(value, str) and value == "None":
                            LOGGER.warning(
                                "String 'None' found in field %s.%s, fixing to NoneType.",
                                table_meta.name(),
                                field_name,
                            )
                            setattr(self, field_name, None)
                            value = None

                        if (
                            qry_data is None
                            and (not field.allowNull() or value)
                            and not ignore_foreignkey
                        ):
                            self._error_manager(
                                "_check_integrity",
                                "INTEGRITY::Relation %s.%s M1 %s.%s with value '%s' type(%s) is invalid"
                                % (
                                    table_meta.name(),
                                    field_name,
                                    relation_m1.foreignTable(),
                                    relation_m1.foreignField(),
                                    value,
                                    type(value),
                                ),
                                self,
                            )

                    elif not field.allowNull():
                        self._error_manager(
                            "_check_integrity",
                            "INTEGRITY::Relationed table %s.%s M1 %s.%s is invalid"
                            % (
                                table_meta.name(),
                                field_name,
                                relation_m1.foreignTable(),
                                relation_m1.foreignField(),
                            ),
                            self,
                        )

        return True

    def relationM1(self, field_name: str = "") -> Optional[Callable]:
        """Return relationM1 object if exists."""

        ret_ = None
        if field_name:
            meta = self.table_metadata().field(field_name)
            if meta is not None:
                meta_rel = meta.relationM1()
                if meta_rel is not None:
                    foreign_table_class = qsadictmodules.QSADictModules.orm_(
                        meta_rel.foreignTable()
                    )
                    if foreign_table_class is not None:
                        foreign_field_obj = getattr(foreign_table_class, meta_rel.foreignField())

                        ret_ = (
                            self._session.query(  # type: ignore [union-attr] # noqa: F821
                                foreign_table_class
                            )
                            .filter(foreign_field_obj == getattr(self, field_name))
                            .first()
                        )
        return ret_

    def relation1M(self, field_name: str = "") -> Dict[str, List[Callable]]:
        """Return relationed instances."""
        ret_ = {}
        field_metadata = self.table_metadata().field(field_name)
        if field_metadata is not None:
            relation_list = field_metadata.relationList()
            for relation in relation_list:
                if relation.cardinality() == pnrelationmetadata.PNRelationMetaData.RELATION_M1:
                    continue

                ft_class = qsadictmodules.QSADictModules.orm_(relation.foreignTable())
                if ft_class is not None:
                    ff_obj = getattr(ft_class, relation.foreignField(), None)
                    if ff_obj is not None:
                        list_ = (
                            ft_class.query(
                                self._session._conn_name  # type: ignore [union-attr] # noqa: F821
                            )
                            .filter(ff_obj == getattr(self, field_name))
                            .all()
                        )
                        ret_["%s_%s" % (relation.foreignTable(), relation.foreignField())] = list_

        else:
            LOGGER.warning("RELATION_1M: invalid field_name %s", field_name)

        return ret_

    def get_transaction_level(self) -> int:
        """Return current transaction level."""

        ret_ = -1
        current_transaction = None
        if self._session:
            if self._session.in_nested_transaction():  # type: ignore [attr-defined]
                current_transaction = self._session.get_nested_transaction()  # type: ignore [attr-defined]
            elif self._session.in_transaction():  # type: ignore [attr-defined]
                current_transaction = self._session.get_transaction()  # type: ignore [attr-defined]

        while True:
            if current_transaction is None:
                break

            ret_ += 1

            current_transaction = current_transaction.parent

        return ret_

    def set_session(self, session: "orm.session.Session") -> None:
        """Set instance session."""
        # LOGGER.warning("Set session %s to instance %s", session, self)
        if not hasattr(self, "_session") or self._session is None:
            # session.add(self)
            self._session = session
        # else:
        #    LOGGER.warning("This instance already belongs to a session")

    def get_session(self) -> Optional["orm.session.Session"]:
        """Get instance session."""

        return self._session

    def get_pk_name(self) -> str:
        """Return primary key."""

        return self.table_metadata().primaryKey()

    def get_pk_value(self) -> Any:
        """Return pk value."""

        return getattr(self, self.get_pk_name())

    def set_pk_value(self, pk_value: Any) -> None:
        """Set pk value."""

        setattr(self, self.get_pk_name(), pk_value)

    def get_mode_access(self) -> int:
        """Return mode_access."""
        if hasattr(self, "_force_mode") and self._force_mode is not None:
            return self._force_mode

        session = self.session

        mode = 3
        if self in session.deleted:
            mode = 2  # delete
        elif self in session.dirty:
            mode = 1  # edit
        elif self._new_object:
            mode = 0  # insert

        if mode in [1, 3] and not self.changes():
            mode = 3

        return mode

    def is_being_created(self) -> bool:
        """Return in being created."""

        return self.mode_access == 0

    def is_being_changed(self) -> bool:
        """Return in being changed."""

        return self.mode_access == 1

    def is_being_deleted(self) -> bool:
        """Return in being deleted."""

        return self.mode_access == 2

    def set_mode_access(self, value: int) -> None:
        """Set forced mode access."""

        self._force_mode = value

    def get_cursor(self) -> "dummy_cursor.DummyCursor":
        """Return dummy cursor."""

        return self._cursor

    def allow_buffer_changed(self, field_name: str, allow: bool = False) -> None:
        """Enable or diable buffer changed signal."""

        if allow:
            if field_name in self._deny_buffer_changed:
                self._deny_buffer_changed.remove(field_name)
        else:
            if field_name not in self._deny_buffer_changed:
                self._deny_buffer_changed.append(field_name)

    def emit_buffer_changed(self, field_name: str) -> None:
        """Emit buffer changed if field is allow."""

        if field_name not in self._deny_buffer_changed:
            # print("EMITE! ", field_name)

            self.bufferChanged.emit(field_name)

    @classmethod
    def _changes_slot(  # type: ignore [attr-defined] # noqa: F821
        cls,
        target,
        new_value: Any,
        old_value: Any,
        event: "orm.attributes.Event",  # type: ignore [name-defined] # noqa: F821
    ) -> None:
        """Change slot."""

        if hasattr(target, "_deny_buffer_changed"):
            # Si no hay funciones conectadas , me voy.
            if not hasattr(target, "bufferChanged") or not target.bufferChanged._remote_funcs:
                return

            # if not hasattr(target, "_cached_bufferchanged"):
            #    target._cached_bufferchanged = {}

            # Si estoy metiendo el mismo valor que tengo cacheado en este campo, no lanzo de nuevo bufferChanged
            if (
                event.key in target._cached_bufferchanged.keys()
                and target._cached_bufferchanged[event.key] == new_value
            ):
                return

            target._cached_bufferchanged[event.key] = new_value
            target.emit_buffer_changed(event.key)
            if event.key in target._cached_bufferchanged.keys():
                del target._cached_bufferchanged[event.key]

    @classmethod
    def _error_manager(cls, text: str, error: Union[Exception, str], obj: object = None) -> None:
        """Return custom error message."""

        exception_: Any = None

        if isinstance(error, str):
            exception_ = Exception
            error_message = error

        else:
            error_info = sys.exc_info()
            exception_ = error_info[0]
            error_message = str(error_info[1])

        LOGGER.error("%s.%s:: %s", cls.__name__, text, error_message, stack_info=False)
        raise exception_(error_message)

    def get_module_iface(self) -> Optional[types.ModuleType]:
        """Return module iface."""
        if not hasattr(
            self, "_action"
        ):  # Puede ocurrir por ejemplo en resultado de consulta de delete cascade
            self._from_query_init(self)

        action = self._action
        if action is not None:
            module_action = None
            if (
                action._mod is not None
                and action._mod.module_name in application.PROJECT.actions.keys()
            ):
                module_action = application.PROJECT.actions[action._mod.module_name]

                if module_action is not None:
                    module_script = module_action._master_widget
                    if module_script is None:
                        module_script = (
                            module_action.load_master_widget()  # type: ignore [unreachable] # noqa: F821
                        )

                    return getattr(module_script, "iface", module_script)

        return None

    def init_serial(self) -> None:
        """Initialice serial field."""
        if (
            self.type(self.pk_name) == "serial"  # noqa E721
            and getattr(self, self.pk_name, None) is None
            and self._session is not None
        ):
            setattr(
                self,
                self.pk_name,
                application.PROJECT.conn_manager.useConn(
                    self._session._conn_name  # type: ignore [attr-defined] # noqa: F821
                )
                .driver()
                .nextSerialVal(self.__tablename__, self.pk_name),
            )

    def init_counter(self, field_name: str) -> None:
        """Initialice counter field."""

        if getattr(self, field_name, None) is None and self._session is not None:
            context_ = getattr(self._action, "_record_widget", None)
            iface = getattr(context_, "iface", None) if context_ else None
            func = getattr(iface, "calculateCounter", None) if iface else None
            value = None
            if func:
                value = func()
            else:
                from pineboolib.application.database import utils

                value = utils.next_counter(field_name, self.cursor)
            if value is not None:
                setattr(self, field_name, value)

    session = property(get_session, set_session)
    transaction_level = property(get_transaction_level)
    pk_name = property(get_pk_name)
    pk = property(get_pk_value, set_pk_value)
    mode_access = property(get_mode_access, set_mode_access)
    cursor = property(get_cursor)
    module_iface = property(get_module_iface)



# -*- coding: utf-8 -*-
"""
PNUnpacker package.

Extract the files from the .abanq and .eneboopkg packages and save them in the flfiles table
"""

from PyQt6 import QtCore  # type: ignore
from typing import Any, List
from pineboolib.core import decorators

err_msgs_: List[str] = []


class PNUnpacker(QtCore.QObject):
    """PNUnpacker Class."""

    def __init__(self, in_: str) -> None:
        """
        Initialize the class.

        @param in_. package file name.
        """

        self.file_ = QtCore.QFile(QtCore.QDir.cleanPath(in_))
        if not self.file_.open(QtCore.QIODevice.OpenModeFlag.ReadOnly):
            raise Exception("Error opening file %r" % in_)
        self.stream_ = QtCore.QDataStream(self.file_)
        self.package_version_ = self.stream_.readBytes().decode("utf-8")

    @decorators.not_implemented_warn
    def errorMessages(self) -> list:
        """
        Return a list of messages with errors that have occurred.

        @return error list.
        """

        return err_msgs_

    def getText(self) -> str:
        """
        Return a record.

        @return record string.
        """

        data_bytes: bytes = QtCore.qUncompress(self.stream_.readBytes()).data()  # type: ignore [call-overload]
        try:
            data_ = data_bytes.decode("utf-8")
        except UnicodeDecodeError:
            data_ = data_bytes.decode("iso-8859-15")

        return data_

    def getBinary(self) -> Any:
        """
        Return a record in byte format.

        @return record bytes.
        """

        return QtCore.qUncompress(self.stream_.readBytes())  # type: ignore [call-overload]

    def getVersion(self) -> str:
        """
        Return the package version.

        @return package version string.
        """

        return self.package_version_

    def jump(self) -> None:
        """
        Skip a field in the package structure.
        """
        self.stream_.readBytes()



"""Translate module."""

from PyQt6 import QtWidgets


def translate(group: str, context: str) -> str:
    """Return the translation if it exists."""

    return QtWidgets.QApplication.translate(group, context)



"""Decorators module."""

from pineboolib.core.utils import logging, utils_base
from pineboolib import application
from pineboolib.qsa import utils


from typing import Callable, Any, TypeVar, cast, Optional

import threading
import functools
import traceback
import time
from sqlalchemy import exc  # type: ignore [import]

TYPEFN = TypeVar("TYPEFN", bound=Callable[..., Any])

LOGGER = logging.get_logger(__name__)


def atomic(
    conn_name: str = "default", wait: bool = True, after_commit: Optional[Callable] = None
) -> "TYPEFN":  # type: ignore [type-var, misc]
    """Return pineboo atomic decorator."""

    def decorator(fun_: TYPEFN) -> TYPEFN:
        @functools.wraps(fun_)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            key = utils_base.session_id(conn_name)
            if wait:
                _wait(key)

            mng_ = application.PROJECT.conn_manager
            while True:
                new_session = utils.driver_session(conn_name)
                if mng_.check_connections():
                    break

            mng_.useConn(conn_name)._session_atomic = new_session
            result_ = None
            try:
                try:
                    with new_session.begin():
                        session_id = utils_base.session_id()
                        orig_transaction = new_session.transaction
                        LOGGER.debug(
                            "New atomic session : %s, connection : %s, transaction: %s",
                            new_session,
                            conn_name,
                            orig_transaction,
                        )

                        try:
                            result_ = fun_(*args, **kwargs)
                            if new_session.transaction is None:
                                LOGGER.warning(
                                    "FIXME:: LA TRANSACCION ATOMICA FINALIZÓ ANTES DE TIEMPO:\nmodule:%s\nfunction:%s\n"
                                    % (fun_.__module__, fun_)
                                )
                            elif new_session.transaction is not orig_transaction:
                                LOGGER.warning(
                                    "FIXME:: LA TRANSACCION ATOMICA FINAL NO ES LA INICIAL:\n module:%s\n function:%s\n"
                                    % (fun_.__module__, fun_)
                                    + "inicial:%s\n final:%s\n"
                                    % (orig_transaction, new_session.transaction)
                                )

                            elif after_commit:
                                # Solo se ejecuta si la transaccion inicial y final es la misma
                                after_commit(session_id)  # type: ignore [truthy-function]
                        except Exception as error:
                            LOGGER.warning(
                                "ATOMIC STACKS\nARGS:%s.\nKWARGS:%s.\nAPP: %s.\nERROR: %s.",
                                args,
                                kwargs,
                                "".join(traceback.format_exc(limit=None)),
                                "".join(traceback.format_stack(limit=None)),
                                stack_info=True,
                            )
                            raise error
                except exc.ResourceClosedError as error:
                    LOGGER.warning("Error al cerrar la transacción : %s, pero continua ....", error)

                _delete_data(conn_name, wait)

            except Exception as error:
                _delete_data(conn_name, wait)
                raise error

            return result_

        mock_fn: TYPEFN = cast(TYPEFN, wrapper)
        return mock_fn

    return decorator  # type: ignore [return-value] # noqa: F723


def serialize(conn_name: str = "default") -> "TYPEFN":  # type: ignore [type-var, misc]
    """Return pineboo atomic decorator."""

    def decorator(fun_: "TYPEFN") -> "TYPEFN":
        @functools.wraps(fun_)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            key = utils_base.session_id(conn_name)
            _wait(key)

            while True:
                if application.PROJECT.conn_manager.check_connections():
                    break

            # new_session = utils.driver_session(conn_name)[1]

            result_ = None
            try:
                LOGGER.debug("New serialize function connection : %s", conn_name)

                try:
                    result_ = fun_(*args, **kwargs)
                except Exception as error:
                    LOGGER.warning(
                        "SERIALIZE STACKS\nARGS:%s.\nKWARGS:%s.\nAPP: %s.\nERROR: %s.",
                        args,
                        kwargs,
                        "".join(traceback.format_exc(limit=None)),
                        "".join(traceback.format_stack(limit=None)),
                        stack_info=True,
                    )
                    raise error

                _delete_data(conn_name)

            except Exception as error:
                _delete_data(conn_name)
                raise error

            return result_

        mock_fn: TYPEFN = cast(TYPEFN, wrapper)
        return mock_fn

    return decorator  # type: ignore [return-value] # noqa: F723


def _wait(key: str) -> None:
    id_thread = threading.current_thread().ident
    if id_thread not in application.SERIALIZE_LIST.keys():
        application.SERIALIZE_LIST[id_thread] = []  # type: ignore [index] # noqa: F821

    application.SERIALIZE_LIST[id_thread].append(key)  # type: ignore [index] # noqa: F821

    while (
        application.SERIALIZE_LIST[id_thread][0] != key  # type: ignore [index] # noqa: F821
    ):  # type: ignore [index] # noqa: F821
        time.sleep(0.01)


def _delete_data(conn_name: str = "", wait: bool = True) -> None:
    """Delete data."""
    mng_ = application.PROJECT.conn_manager
    conn = mng_.useConn(conn_name)

    if conn is not None:
        if application.SHOW_CONNECTION_EVENTS:
            LOGGER.debug("Removing sessions from connection %s", conn_name)
        if conn._session_atomic is not None:
            conn._session_atomic.close()
            conn._session_atomic = None
        if conn._session_legacy is not None:
            conn._session_legacy.close()
            conn._session_legacy = None

    if mng_.REMOVE_CONNECTIONS_AFTER_ATOMIC:
        time.sleep(0.05)
        for item in mng_.enumerate():
            if application.SHOW_CONNECTION_EVENTS:
                LOGGER.debug("Removing connection %s after decorator", item)
            mng_.removeConn(item)

    if wait:
        id_thread: int = threading.current_thread().ident or -1
        key = utils_base.session_id(conn_name)
        if id_thread in application.SERIALIZE_LIST.keys():
            if key in application.SERIALIZE_LIST[id_thread]:  # type: ignore [index] # noqa: F821
                application.SERIALIZE_LIST[id_thread].remove(
                    key
                )  # type: ignore [index] # noqa: F821



# -*- coding: utf-8 -*-
"""
AQUtil Module.

Use the resources of pineboolib.fllegacy.flutil.FLUtil.
"""

from pineboolib import logging
from pineboolib.fllegacy import flutil

LOGGER = logging.get_logger(__name__)


class AQUtil(flutil.FLUtil):
    """AQUtil Class."""

    def __init__(self) -> None:
        """Initialize a new instance."""



"""
SysBaseType for QSA.

Will be inherited at fllegacy.
"""

import platform
import traceback
import ast

from typing import Any, Dict, Optional, List, Union

from PyQt6 import QtWidgets, QtXml, QtCore

from pineboolib.core import settings, decorators
from pineboolib.core.utils.utils_base import ustr, filedir
from pineboolib.core.utils import logging

from pineboolib import application
from pineboolib.application import connections

from pineboolib.application.process import Process

LOGGER = logging.get_logger(__name__)


class SysBaseType(object):
    """
    Obtain useful data from the application.
    """

    @classmethod
    def nameUser(cls) -> str:
        """Get current database user."""

        ret_ = (
            application.PROJECT.DGI.get_nameuser()
            if application.PROJECT.DGI.use_alternative_credentials()
            else application.PROJECT.conn_manager.mainConn().user()
        )

        return ret_ or ""

    # @classmethod
    # def interactiveGUI(cls) -> str:
    #    """Check if running in GUI mode."""
    #    return application.PROJECT.DGI.interactiveGUI()

    @classmethod
    def isUserBuild(cls) -> bool:
        """Check if this build is an user build."""
        return cls.version().upper().find("USER") > -1

    @classmethod
    def isDeveloperBuild(cls) -> bool:
        """Check if this build is a developer build."""
        return cls.version().upper().find("DEVELOPER") > -1

    @classmethod
    def isNebulaBuild(cls) -> bool:
        """Check if this build is a nebula build."""
        return cls.version().upper().find("NEBULA") > -1

    @classmethod
    def isDebuggerMode(cls) -> bool:
        """Check if running in debugger mode."""
        return bool(settings.CONFIG.value("application/isDebuggerMode", False))

    @classmethod
    @decorators.not_implemented_warn
    def isCloudMode(cls) -> bool:
        """Check if running on cloud mode."""
        return False

    @classmethod
    def isDebuggerEnabled(cls) -> bool:
        """Check if this debugger is on."""
        return bool(application.PROJECT.db_admin_mode)

    @classmethod
    def isQuickBuild(cls) -> bool:
        """Check if this build is a Quick build."""
        return not cls.isDebuggerEnabled()

    @classmethod
    def isLoadedModule(cls, modulename: str) -> bool:
        """Check if a module has been loaded."""
        return modulename in application.PROJECT.conn_manager.managerModules().listAllIdModules()

    @classmethod
    def osName(cls) -> str:
        """
        Get operating system name.

        @return Código del sistema operativo (WIN32, LINUX, MACX)
        """
        if platform.system() == "Windows":
            return "WIN32"
        elif platform.system() in ["Linux", "Linux2"]:
            return "LINUX"
        elif platform.system() == "Darwin":
            return "MACX"
        else:
            return platform.system()

    @classmethod
    def nameBD(cls) -> str:
        """Get database name."""
        return application.PROJECT.conn_manager.mainConn().DBName()

    @classmethod
    def toUnicode(cls, val: str, format_encode: str) -> str:
        """Convert string to unicode."""
        return val.encode(format_encode, "replace").decode("utf-8", "replace")

    @classmethod
    def fromUnicode(cls, val: str, format_encode: str) -> str:
        """Convert from unicode to string."""
        return val.encode("utf-8", "replace").decode(format_encode, "replace")

    @classmethod
    def Mr_Proper(cls) -> None:
        """Cleanup database like Mr. Proper."""
        application.PROJECT.conn_manager.mainConn().Mr_Proper()

    @classmethod
    def installPrefix(cls) -> str:
        """Get folder where app is installed."""
        return filedir("..")

    @classmethod
    def version(cls) -> str:
        """Get version number as string."""
        return str(application.PROJECT.load_version())

    @classmethod
    def processEvents(cls) -> None:
        """Process event loop."""
        QtWidgets.QApplication.processEvents()

    @classmethod
    def write(cls, encode_: str, dir_: str, contenido: str) -> None:
        """Write to file."""
        from pineboolib.application.types import File

        file_iso = File(dir_, encode_)
        file_iso.write(contenido)
        file_iso.close()

    @classmethod
    def cleanupMetaData(cls, conn_name: str = "default") -> None:
        """Clean up metadata."""
        application.PROJECT.conn_manager.manager().cleanupMetaData()

    @classmethod
    def nameDriver(cls, conn_name: str = "default") -> str:
        """Get driver name."""
        return application.PROJECT.conn_manager.useConn(conn_name).driverName()

    @classmethod
    def nameHost(cls, conn_name: str = "default") -> str:
        """Get database host name."""
        return application.PROJECT.conn_manager.useConn(conn_name).host()

    @classmethod
    def addDatabase(
        cls,
        driver_or_conn: str,
        db_name: str = "",
        db_user: str = "",
        db_pw: str = "",
        db_host: str = "",
        db_port: int = 0,
        conn_name: str = "",
    ) -> bool:
        """Add a new database."""

        mng_ = application.PROJECT.conn_manager
        conn_db = None
        new_conn = False

        if not db_name:
            conn_db = mng_.useConn(driver_or_conn)
        else:
            if not conn_name:
                raise Exception(
                    "Invalid connection data. conn_name: %s, driver_name: %s, database_name: %s, user_name: %s, db_host: %s, port: %s"
                    % (conn_name, driver_or_conn, db_name, db_user, db_host, db_port)
                )
            new_conn = True
            conn_db = mng_.useConn(conn_name, db_name)

        if not conn_db.isOpen():
            conn_db_name = db_name
            conn_db_host = db_host
            conn_db_port = db_port
            conn_db_user = db_user
            conn_db_pass = db_pw

            if new_conn:
                conn_db._driver_name = driver_or_conn.lower()
            else:
                main_conn = mng_.mainConn()
                conn_db_name = main_conn._db_name
                conn_db_host = main_conn._db_host
                conn_db_port = main_conn._db_port
                conn_db_user = main_conn._db_user_name
                conn_db_pass = main_conn._db_password

            if mng_._drivers_sql_manager.loadDriver(conn_db._driver_name):
                conn_db.conn = conn_db.conectar(
                    conn_db_name, conn_db_host, conn_db_port, conn_db_user, conn_db_pass
                )

                if isinstance(conn_db.conn, bool):
                    return False

                conn_db._is_open = True

        else:
            LOGGER.debug(
                "addDatabase: '%s' connection is already open",
                conn_name if conn_name else driver_or_conn,
            )

        return True

    @classmethod
    def removeDatabase(cls, conn_name: str = "default") -> bool:
        """Remove a database."""
        return application.PROJECT.conn_manager.removeConn(conn_name)

    @classmethod
    def idSession(cls) -> str:
        """Get Session ID."""

        # FIXME: Code copied from flapplication.aqApp
        return application.ID_SESSION

    @classmethod
    def reportChanges(cls, changes: Dict[str, str] = {}):
        """Create a report for project changes."""
        ret = ""
        # DEBUG:: FOR-IN: ['key', 'changes']
        for key in changes.keys():
            if key == "size":
                continue
            chg = changes[key].split("@")
            ret += "Nombre: %s \n" % chg[0]
            ret += "Estado: %s \n" % chg[1]
            ret += "ShaOldTxt: %s \n" % chg[2]
            ret += "ShaNewTxt: %s \n" % chg[3]
            ret += "###########################################\n"

        return ret

    @classmethod
    def diffXmlFilesDef(
        cls, xml_old: "QtXml.QDomNode", xml_new: "QtXml.QDomNode"
    ) -> Dict[str, Any]:
        """Create a Diff for XML."""
        array_old = cls.filesDefToArray(xml_old)
        array_new = cls.filesDefToArray(xml_new)
        ret: Dict[str, Any] = {}
        size = 0
        # DEBUG:: FOR-IN: ['key', 'array_old']
        for key in array_old:
            if key not in array_new:
                info = [key, "del", array_old[key]["shatext"], array_old[key]["shabinary"], "", ""]
                ret[key] = "@".join(info)
                size += 1
        # DEBUG:: FOR-IN: ['key', 'array_new']

        for key in array_new:
            if key not in array_old:
                info = [key, "new", "", "", array_new[key]["shatext"], array_new[key]["shabinary"]]
                ret[key] = "@".join(info)
                size += 1
            else:
                if (
                    array_new[key]["shatext"] != array_old[key]["shatext"]
                    or array_new[key]["shabinary"] != array_old[key]["shabinary"]
                ):
                    info = [
                        key,
                        "mod",
                        array_old[key]["shatext"],
                        array_old[key]["shabinary"],
                        array_new[key]["shatext"],
                        array_new[key]["shabinary"],
                    ]
                    ret[key] = "@".join(info)
                    size += 1

        ret["size"] = size
        return ret

    @classmethod
    def filesDefToArray(cls, xml: "QtXml.QDomNode") -> Dict[str, Dict[str, str]]:
        """Convert Module MOD xml to array."""
        root = xml.firstChild()
        files = root.childNodes()
        ret = {}
        i = 0
        while i < len(files):
            item = files.item(i)
            fil = {
                "id": item.namedItem("name").toElement().text(),
                "module": item.namedItem("module").toElement().text(),
                "text": item.namedItem("text").toElement().text(),
                "shatext": item.namedItem("shatext").toElement().text(),
                "binary": item.namedItem("binary").toElement().text(),
                "shabinary": item.namedItem("shabinary").toElement().text(),
            }
            i += 1
            if len(fil["id"]) == 0:
                continue
            ret[fil["id"]] = fil

        return ret

    @classmethod
    def textPacking(cls, ext: str = "") -> bool:
        """Determine if file is text."""
        return ext.endswith(
            (
                ".ui",
                ".qry",
                ".kut",
                ".jrxml",
                ".ar",
                ".mtd",
                ".ts",
                ".qs",
                ".py",
                ".xml",
                ".xpm",
                ".svg",
            )
        )

    @classmethod
    def binaryPacking(cls, ext: str = "") -> bool:
        """Determine if file is binary."""
        return ext.endswith(".qs")

    @classmethod
    def infoMsgBox(cls, msg: str = "") -> None:
        """Show information message box."""
        msg = ustr(msg)
        msg += "\n"
        application.PROJECT.message_manager().send("msgBoxInfo", None, [msg])

    @classmethod
    def warnMsgBox(cls, msg: str = "", *buttons: Any) -> None:
        """Show Warning message box."""
        new_list = []
        new_list.append("%s.\n" % ustr(msg))
        for button in buttons:
            new_list.append(button)

        application.PROJECT.message_manager().send("msgBoxWarning", None, new_list)

    @classmethod
    def errorMsgBox(cls, msg: Optional[str] = None) -> None:
        """Show error message box."""
        msg = ustr(msg)
        msg += "\n"

        application.PROJECT.message_manager().send("msgBoxError", None, [msg])

    @classmethod
    def translate(cls, text: str, arg2: Optional[str] = None) -> str:
        """Translate text."""
        if arg2:
            return arg2
        return text

    @classmethod
    def _warnHtmlPopup(cls, html: str, options: List) -> None:
        """Show a fancy html popup."""
        raise Exception("not implemented.")

    @classmethod
    def infoPopup(cls, msg: Optional[str] = None) -> None:
        """Show information popup."""
        msg = ustr(msg)
        caption = cls.translate("AbanQ Información")
        msg = msg.replace("\n", "<br>")
        msg_html = ustr(
            '<img source="about.png" align="right">',
            "<b><u>",
            caption,
            "</u></b><br><br>",
            msg,
            "<br>",
        )
        cls._warnHtmlPopup(msg_html, [])

    @classmethod
    def warnPopup(cls, msg: str = "") -> None:
        """Show warning popup."""
        msg = ustr(msg)
        msg = msg.replace("\n", "<br>")
        caption = cls.translate("AbanQ Aviso")
        msg_html = ustr(
            '<img source="bug.png" align="right">',
            "<b><u>",
            caption,
            "</u></b><br><br>",
            msg,
            "<br>",
        )
        cls._warnHtmlPopup(msg_html, [])

    @classmethod
    def errorPopup(cls, msg: str = "") -> None:
        """Show error popup."""
        msg = ustr(msg)
        msg = msg.replace("\n", "<br>")
        caption = cls.translate("AbanQ Error")
        msg_html = ustr(
            '<img source="remove.png" align="right">',
            "<b><u>",
            caption,
            "</u></b><br><br>",
            msg,
            "<br>",
        )
        cls._warnHtmlPopup(msg_html, [])

    @classmethod
    def trTagText(cls, tag_text: str = "") -> str:
        """Process QT_TRANSLATE_NOOP tags."""
        if not tag_text.startswith("QT_TRANSLATE_NOOP"):
            return tag_text
        txt = tag_text[len("QT_TRANSLATE_NOOP") + 1 :]
        txt = "[%s]" % txt[0 : len(txt) - 1]
        arr = ast.literal_eval(txt)  # FIXME: Don't use "ast.literal_eval"
        return cls.translate(arr[0], arr[1])

    @classmethod
    def updatePineboo(cls) -> None:
        """Execute auto-updater."""
        QtWidgets.QMessageBox.warning(
            QtWidgets.QApplication.focusWidget(),
            "Pineboo",
            cls.translate("Funcionalidad no soportada aún en Pineboo."),
            QtWidgets.QMessageBox.StandardButton.Ok,
        )
        return

    @classmethod
    def setObjText(cls, container: "QtWidgets.QWidget", component: str, value: str = "") -> bool:
        """Set text to random widget."""
        object_ = cls.testObj(container, component)
        if object_ is None:
            return False
        clase = object_.__class__.__name__

        if clase in ["QPushButton", "QToolButton"]:
            pass
        elif clase == "QLabel":
            cls.runObjMethod(container, component, "text", value)
        elif clase == "FLFieldDB":
            cls.runObjMethod(container, component, "setValue", value)
        else:
            return False
        return True

    @classmethod
    def disableObj(cls, container: "QtWidgets.QWidget", component: str) -> bool:
        """Disable random widget."""
        object_ = cls.testObj(container, component)
        if not object_:
            return False

        clase = object_.__class__.__name__

        if clase in ["QToolButton", "QPushButton"]:
            cls.runObjMethod(container, component, "setEnabled", False)
        elif clase == "FLFieldDB":
            cls.runObjMethod(container, component, "setDisabled", True)
        else:
            return False

        return True

    @classmethod
    def enableObj(cls, container: "QtWidgets.QWidget", component: str) -> bool:
        """Enable random widget."""
        object_ = cls.testObj(container, component)
        if not object_:
            return False

        clase = object_.__class__.__name__

        if clase == "QPushButton":
            pass
        elif clase == "QToolButton":
            cls.runObjMethod(container, component, "setEnabled", True)
        elif clase == "FLFieldDB":
            cls.runObjMethod(container, component, "setDisabled", False)
        else:
            return False
        return True

    @classmethod
    def filterObj(cls, container: QtWidgets.QWidget, component: str, filter: str = "") -> bool:
        """Apply filter to random widget."""
        object_ = cls.testObj(container, component)
        if object_ is None:
            return False

        clase = object_.__class__.__name__

        if clase == "FLTableDB":
            pass
        elif clase == "FLFieldDB":
            cls.runObjMethod(container, component, "setFilter", filter)
        else:
            return False

        return True

    @classmethod
    def testObj(
        cls, container: Optional["QtWidgets.QWidget"] = None, component: str = ""
    ) -> Optional[Union["QtCore.QObject", "QtWidgets.QWidget"]]:
        """Test if object does exist."""

        object_ = None
        if container is not None:
            object_ = container.findChild(QtWidgets.QWidget, component)
            if object_ is None:
                child_fun = getattr(container, "child", None)  # type: ignore [unreachable]
                if child_fun is not None:
                    object_ = child_fun(component)

            if object_ is None:
                LOGGER.warning(  # type: ignore [unreachable]
                    "%s no existe en %s", component, container
                )
        return object_

    @classmethod
    def testAndRun(
        cls, container: "QtWidgets.QWidget", component: str, method: str = "", param: Any = None
    ) -> bool:
        """Test and execute object."""
        object_ = cls.testObj(container, component)
        if not object_:
            return False
        if not cls.runObjMethod(container, component, method, param):
            return False
        return True

    @classmethod
    def runObjMethod(
        cls, container: "QtWidgets.QWidget", component: str, method: str, param: Any = None
    ) -> bool:
        """Execute method from object."""
        object_ = cls.testObj(container, component)
        attr = getattr(object_, method, None)
        if attr is not None:
            attr(param)
        else:
            LOGGER.warning("%s no existe en %s", method, component)

        return True

    @classmethod
    def connectSS(
        cls, sender: "QtWidgets.QWidget", signal: str, receiver: "QtWidgets.QWidget", slot: str
    ) -> bool:
        """Connect signal to slot."""
        if not sender:
            return False
        connections.connect(sender, signal, receiver, slot)
        return True

    @classmethod
    def openUrl(cls, url: Union[str, List[str]] = "") -> bool:
        """Open given URL in a browser."""
        if not url:
            return False

        if isinstance(url, List):
            url = url[0]

        os_name = cls.osName()
        if os_name == "LINUX":
            if cls.launchCommand(["xdg-open", url]):
                return True
            elif cls.launchCommand(["gnome-open", url]):
                return True
            elif cls.launchCommand(["kfmclient openURL", url]):
                return True
            elif cls.launchCommand(["kfmclient exec", url]):
                return True
            elif cls.launchCommand(["firefox", url]):
                return True
            elif cls.launchCommand(["mozilla", url]):
                return True
            elif cls.launchCommand(["opera", url]):
                return True
            elif cls.launchCommand(["google-chrome", url]):
                return True
            return False

        elif os_name == "WIN32":
            if url.startswith("mailto"):
                url = url.replace("&", "^&")
            return cls.launchCommand(["cmd.exe", "/C", "start", "", url])

        elif os_name == "MACX":
            return cls.launchCommand(["open", url])

        return False

    @classmethod
    def launchCommand(cls, comando: Union[str, List[str]]) -> bool:
        """Execute a program."""
        try:
            proc = Process()
            proc.execute(comando)
            return True
        except Exception:
            LOGGER.error(traceback.format_exc())
            return False



"""
QSADictModules.

Manages read and writting QSA dynamic properties that are loaded during project startup.
"""

from pineboolib.application import xmlaction, proxy, safeqsa
from pineboolib import logging, application

import sqlalchemy  # type: ignore [import]
import threading
import gc

from typing import Any, Optional, Union, TYPE_CHECKING

LOGGER = logging.get_logger(__name__)

if TYPE_CHECKING:
    from pineboolib.application.database.orm import basemodel


class QSADictModulesThread:
    pass


class QSADictModules:
    """
    Manage read and write dynamic properties for QSA.
    """

    _qsa_dict_modules = None
    _tree_thread = {}

    @classmethod
    def qsa_dict_modules(cls) -> Any:
        """Retrieve QSA module, hidding it from MyPy."""
        if cls._qsa_dict_modules is None:
            if TYPE_CHECKING:
                qsa_dict_modules_tree: Any = None  # pragma: no cover
            else:
                from pineboolib.application import modules_tree as qsa_dict_modules_tree

            cls._qsa_dict_modules = qsa_dict_modules_tree
        return cls._qsa_dict_modules

    @classmethod
    def from_project(cls, script_name: str) -> Any:
        """
        Return project object for given name.
        """
        module_name = "sys_module" if script_name == "sys" else script_name

        id_thread = threading.current_thread().ident
        if not id_thread in cls._tree_thread.keys():
            cls._tree_thread[id_thread] = QSADictModulesThread()

        if not hasattr(cls._tree_thread[id_thread], module_name):
            LOGGER.warning("Creando %s en el hilo %d" % (module_name, id_thread))
            ret_original = getattr(cls.qsa_dict_modules(), module_name, None)
            setattr(cls._tree_thread[id_thread], module_name, ret_original)

        ret_ = getattr(cls._tree_thread[id_thread], module_name, None)
        if ret_ is None and not module_name.endswith("orm"):
            LOGGER.warning("Module %s not found!", module_name)

        return ret_

    @classmethod
    def orm_(cls, action_name: str = "", show_error: bool = True) -> Any:
        """Return orm instance."""

        table_name = (
            application.PROJECT.actions[action_name]._table
            if action_name in application.PROJECT.actions.keys()
            else action_name
        )

        if table_name:
            orm = cls.from_project("%s_orm" % (table_name))
            if orm is not None:
                init_fn = getattr(orm, "_qsa_init", None)
                if init_fn:
                    sqlalchemy.event.listen(orm, "init", init_fn)

                return orm
            else:
                if show_error:
                    LOGGER.error("Model %s not found!", table_name, stack_info=True)

        return None

    @classmethod
    def action_exists(cls, scriptname: str) -> bool:
        """
        Check if action is already loaded.
        """
        return hasattr(cls.qsa_dict_modules(), scriptname)

    @classmethod
    def set_qsa_tree(
        cls,
        script_name: str,
        action_or_model: Optional[Union["proxy.DelayedObjectProxyLoader", "basemodel.BaseModel"]],
    ) -> None:
        """
        Save action or other objects for QSA.
        """
        setattr(cls.qsa_dict_modules(), script_name, action_or_model)

    @classmethod
    def save_action_for_root_module(cls, action: "xmlaction.XMLAction") -> bool:
        """Save a new module as an action."""

        module_name = action._name if action._name != "sys" else "sys_module"
        if cls.action_exists(module_name):
            if module_name != "sys_module":
                LOGGER.info("Module found twice, will not be overriden: %s", module_name)
            return False

        # Se crea la action del módulo
        delayed_action = proxy.DelayedObjectProxyLoader(
            action.load_master_widget, name="QSA.Module.%s" % module_name
        )
        cls.set_qsa_tree(module_name, delayed_action)
        safeqsa.SafeQSA.save_root_module(module_name, delayed_action)
        return True

    @classmethod
    def save_action_for_mainform(cls, action: "xmlaction.XMLAction"):
        """Save a new mainform."""

        name = action._name
        module = action._mod
        if module is None:
            raise ValueError("Action.module must be set before calling")

        actionname = "form%s" % name
        if cls.action_exists(actionname):
            LOGGER.info(
                "No se sobreescribe variable de entorno %s. Hay una definición previa.",
                "%s.form%s" % (module.module_name, name),
            )
            return False
        # Se crea la action del form
        delayed_action = proxy.DelayedObjectProxyLoader(
            action.load_master_widget, name="QSA.Module.%s.Action.form%s" % (module.mod.name, name)
        )
        cls.set_qsa_tree(actionname, delayed_action)
        safeqsa.SafeQSA.save_mainform(actionname, delayed_action)
        return True

    @classmethod
    def save_action_for_formrecord(cls, action: "xmlaction.XMLAction"):
        """Save a new formrecord."""
        name = action._name
        module = action._mod
        if module is None:
            raise ValueError("Action.module must be set before calling")
        actionname = "formRecord" + name
        if cls.action_exists(actionname):
            LOGGER.info(
                "No se sobreescribe variable de entorno %s. Hay una definición previa.",
                "%s.formRecord%s" % (module.module_name, name),
            )
            return False
        # Se crea la action del formRecord
        delayed_action = proxy.DelayedObjectProxyLoader(
            action.load_record_widget,
            name="QSA.Module.%s.Action.formRecord%s" % (module.mod.name, name),
        )

        cls.set_qsa_tree(actionname, delayed_action)
        safeqsa.SafeQSA.save_formrecord(actionname, delayed_action)
        return True

    @classmethod
    def save_action_for_class(cls, action: "xmlaction.XMLAction"):
        """Save action class action."""

        class_name = action._class_script
        module = action._mod
        if module is None:
            raise ValueError("Action.module must be set before calling")

        if class_name:
            if cls.action_exists(class_name):
                LOGGER.info(
                    "No se sobreescribe variable de entorno %s. Hay una definición previa.",
                    "%s.%s" % (module.module_name, class_name),
                )
                return False

            delayed_action = proxy.DelayedObjectProxyLoader(
                action.load_class,
                name="QSA.Module.%s.Action.class_%s" % (module.mod.name, class_name),
            )
            cls.set_qsa_tree(action._name, delayed_action)

    @classmethod
    def clean_all(cls):
        """Clean all saved data."""
        qsa_dict_modules = cls.qsa_dict_modules()

        safeqsa.SafeQSA.clean_all()
        for name in [attr for attr in dir(qsa_dict_modules) if not attr[0] == "_"]:
            att = getattr(qsa_dict_modules, name)
            if isinstance(att, proxy.DelayedObjectProxyLoader) or (
                name.endswith(("_orm", "_class"))
                and (not name.startswith("fl") and name not in ("flusers", "flgroups"))
            ):
                delattr(qsa_dict_modules, name)

        gc.collect()


from_project = QSADictModules.from_project
orm_ = QSADictModules.orm_



# -*- coding: utf-8 -*-
"""
Defines the IConnection class.
"""

from pineboolib.core.utils import logging


from typing import Any, List, Dict, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.fllegacy import flmanager  # pragma: no cover
    from pineboolib.fllegacy import flmanagermodules  # pragma: no cover
    from pineboolib.application.metadata import pntablemetadata  # pragma: no cover
    from pineboolib.application.database import (
        pnsqldriversmanager,
        pnconnectionmanager,
    )  # pragma: no cover
    from pineboolib.interfaces import isession, isqldriver, isqlcursor  # pragma: no cover
    from sqlalchemy.engine import base, result  # type: ignore [import]

LOGGER = logging.get_logger(__name__)


class IConnection:
    """Interface for database cursors which are used to emulate FLSqlCursor."""

    _db_name: str
    _db_host: str
    _db_port: int
    _db_user_name: str
    _db_password: str = ""
    conn: Optional[Union["base.Connection", bool]]
    _interactive_gui: str
    _drivers_sql_manager: "pnsqldriversmanager.PNSqlDriversManager"
    _driver_name: str
    _name: str
    _is_open: bool
    _last_error: str
    _transaction_level: int
    _conn_manager: "pnconnectionmanager.PNConnectionManager"
    _driver: Optional["isqldriver.ISqlDriver"]
    _last_active_cursor: Optional["isqlcursor.ISqlCursor"]
    _session_legacy: Optional["isession.PinebooSession"]
    _session_atomic: Optional["isession.PinebooSession"]

    def connectionName(self) -> str:
        """Get the current connection name for this cursor."""
        return ""  # pragma: no cover

    def isOpen(self) -> bool:
        """Indicate if a connection is open."""
        return False  # pragma: no cover

    def tables(self, tables: Union[str, int] = "") -> List[str]:
        """Return a list of available tables in the database, according to a given filter."""
        return []  # pragma: no cover

    def DBName(self) -> str:
        """Return the database name."""
        return ""  # pragma: no cover

    def driver(self) -> "isqldriver.ISqlDriver":
        """Return the instance of the driver that is using the connection."""

        return None  # type: ignore [return-value] # pragma: no cover

    def database(self) -> "IConnection":
        """Return self."""

        return self  # pragma: no cover

    def session(self) -> "isession.PinebooSession":
        """
        Sqlalchemy session.

        When using the ORM option this function returns the session for sqlAlchemy.
        """

        return None  # type: ignore[return-value] # noqa: F821 # pragma: no cover

    def connManager(self) -> "pnconnectionmanager.PNConnectionManager":
        """Return connection manager."""

        return None  # type: ignore[return-value] # noqa: F821 # pragma: no cover

    def engine(self) -> "base.Engine":
        """Sqlalchemy connection."""

        return None  # type: ignore[return-value] # noqa: F821 # pragma: no cover

    def dictDatabases(self) -> Dict[str, "IConnection"]:
        """Return dict with own database connections."""

        return {}  # pragma: no cover

    # def cursor(self) -> IApiCursor:
    #    """Return a cursor to the database."""

    #    return IApiCursor()

    def connection(self) -> "base.Connection":  # type: ignore [empty-body]
        """Return base connection."""

        pass  # pragma: no cover

    def lastActiveCursor(self) -> Optional["isqlcursor.ISqlCursor"]:  # returns FLSqlCuror
        """Return the last active cursor in the sql driver."""

        return None  # pragma: no cover

    def conectar(
        self, db_name, db_host, db_port, db_user_name, db_returnword
    ) -> Union["base.Connection", bool]:
        """Request a connection to the database."""

        return False  # pragma: no cover

    def driverName(self) -> str:
        """Return sql driver name."""

        return ""  # pragma: no cover

    def driverAlias(self) -> str:
        """Return sql driver alias."""

        return ""  # pragma: no cover

    def driverNameToDriverAlias(self, name) -> str:
        """Return the alias from the name of a sql driver."""

        return ""  # pragma: no cover

    def lastError(self) -> str:
        """Return the last error reported by the sql driver."""

        return ""  # pragma: no cover

    def host(self) -> str:
        """Return the name of the database host."""

        return ""  # pragma: no cover

    def port(self) -> int:
        """Return the port used by the database."""

        return 0  # pragma: no cover

    def user(self) -> str:
        """Return the user name used by the database."""

        return ""  # pragma: no cover

    def returnword(self) -> str:
        """Return ****word used by the database."""
        return ""  # pragma: no cover

    def md5TuplesStateTable(self, curname: str) -> bool:
        """
        Return the sum md5 with the total records inserted, deleted and modified in the database so far.

        Useful to know if the database has been modified from a given moment.
        """

        return True  # pragma: no cover

    def setInteractiveGUI(self, data) -> None:
        """Set if it is an interactive GUI."""

        return  # pragma: no cover

    def setQsaExceptions(self, data) -> None:
        """See properties of the qsa exceptions."""

        return  # pragma: no cover

    def formatValue(self, type_, value, upper) -> str:
        """Return a correctly formatted value to be assigned as a where filter."""

        return ""  # pragma: no cover

    def formatValueLike(self, type_, value, upper) -> str:
        """Return a correctly formatted value to be assigned as a WHERE LIKE filter."""

        return ""  # pragma: no cover

    def doTransaction(self, cursor) -> bool:
        """Make a transaction or savePoint according to transaction level."""

        return False  # pragma: no cover

    def transactionLevel(self) -> int:
        """Indicate the level of transaction."""

        return 0  # pragma: no cover

    def doRollback(self, cur) -> bool:
        """Drop a transaction or savepoint depending on the transaction level."""

        return False  # pragma: no cover

    def interactiveGUI(self) -> str:
        """Return if it is an interactive GUI."""

        return ""  # pragma: no cover

    def doCommit(self, cur, notify=True) -> bool:
        """Approve changes to a transaction or a save point based on your transaction level."""

        return False  # pragma: no cover

    def canDetectLocks(self) -> bool:
        """Indicate if the connection detects locks in the database."""

        return False  # pragma: no cover

    def commit(self) -> bool:
        """Send the commit order to the database."""

        return True  # pragma: no cover

    def canOverPartition(self) -> bool:
        """Return True if the database supports the OVER statement."""

        return True  # pragma: no cover

    def canRegenTables(self) -> bool:
        """Return True if the database can regenerate tables."""

        return True  # pragma: no cover

    def regenTable(self, table_name: str, mtd: "pntablemetadata.PNTableMetaData") -> bool:
        """Regenerate a table."""

        return True  # pragma: no cover

    def releaseSavePoint(self, save_point: int) -> bool:
        """Release a save point."""

        return True  # pragma: no cover

    def Mr_Proper(self) -> None:
        """Clean the database of unnecessary tables and records."""

        return  # pragma: no cover

    def transaction(self) -> bool:
        """Create a transaction/savePoint."""

        return True  # pragma: no cover

    def rollback(self) -> bool:
        """Roll back a transaction/savepoint."""

        return True  # pragma: no cover

    def nextSerialVal(self, table: str, field: str) -> int:
        """Indicate next available value of a serial type field."""

        return 0  # pragma: no cover

    def existsTable(self, name: str) -> bool:
        """Indicate the existence of a table in the database."""

        return False  # pragma: no cover

    def createTable(self, tmd: "pntablemetadata.PNTableMetaData") -> bool:
        """Create a table in the database, from a pntablemetadata.PNTableMetaData."""

        return False  # pragma: no cover

    def mismatchedTable(self, tablename: str, tmd: "pntablemetadata.PNTableMetaData") -> bool:
        """Compare an existing table with a pntablemetadata.PNTableMetaData and return if there are differences."""

        return False  # pragma: no cover

    def normalizeValue(self, text: str) -> Optional[str]:
        """Return the value of a correctly formatted string to the database type from a string."""

        return None  # pragma: no cover

    # def queryUpdate(self, name: str, update: str, filter: str) -> Optional[str]:
    #    """Return a correct UPDATE query for the database type."""

    #    return ""  # pragma: no cover

    def execute_query(self, query: str) -> Optional["result.Result"]:  # type: ignore [name-defined]
        """Execute a query in a database cursor."""

        return None  # pragma: no cover

    def alterTable(self, table_metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Modify the fields of a table in the database based on the differences of two pntablemetadata.PNTableMetaData."""

        return False  # pragma: no cover

    def getTimeStamp(self) -> str:
        """Return timestamp."""

        return ""  # pragma: no cover

    def manager(self) -> "flmanager.FLManager":
        """Return flmanager instance."""

        return self.connManager().manager()  # pragma: no cover

    def managerModules(self) -> "flmanagermodules.FLManagerModules":
        """Return flmanager instance."""

        return self.connManager().managerModules()  # pragma: no cover

    def singleConnection(self) -> bool:
        """Return if driver uses a single connection."""
        return False  # pragma: no cover

    def sqlLength(self, field_name: str, size: int) -> str:
        """Return length formated."""

        return ""  # pragma: no cover

    def insert_data(self, table_name: str, fields: List[str], values: List[str]) -> bool:
        """Insert data into table."""

        return True  # pragma: no cover

    def update_data(
        self, metadata: "pntablemetadata.PNTableMetaData", data: Dict[str, Any], pk_value: Any
    ) -> bool:
        """Update table data."""

        return True  # pragma: no cover

    def delete_data(self, metadata: "pntablemetadata.PNTableMetaData", pk_value: Any) -> bool:
        """Delete row from table."""

        return True  # pragma: no cover

    def close(self) -> None:
        """Close connection."""

        return  # pragma: no cover

    def resolve_dsn(self) -> str:
        """Return dsn data."""

        return ""



"""AQSHttp module."""

from pineboolib.q3widgets import qhttp


class AQSHttp(qhttp.QHttp):
    """AQSHttp class."""

    pass


class AQSHttpRequestHeader(qhttp.QHttpRequestHeader):
    """AQSHttpRequestHeader class."""

    pass


class AQSHttpResponseHeader(qhttp.QHttpResponseHeader):
    """AQSHttpRequestHeader class."""

    pass



"""Dgi_schema module."""
# -*- coding: utf-8 -*-
from importlib import import_module
from typing import cast, Optional, Any
from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib.application.utils.mobilemode import is_mobile_mode
from pineboolib import logging

LOGGER = logging.get_logger(__name__)


class DgiSchema(object):
    """dgi_schema class."""

    _desktop_enabled: bool
    _ml_default: bool
    _name: str
    _alias: str
    _local_desktop: bool
    _mobile: bool
    _clean_no_python: bool
    # FIXME: Guess this is because there is conditional code we don't want to run on certain DGI
    # .... this is really obscure. Please avoid at all costs. Having __NO_PYTHON__ is bad enough.
    _alternative_content_cached: bool

    def __init__(self) -> None:
        """Inicialize."""

        # FIXME: This init is intended to be called only on certain conditions.
        # ... Worse than it seems: looks like this class is prepared to be constructed without
        # ... calling __init__, on purpose, to have different behavior than calling it.

        self._desktop_enabled = True  # Indica si se usa en formato escritorio con interface Qt
        self._ml_default = True
        self._local_desktop = True
        self._name = "dgi_shema"
        self._alias = "Default Schema"
        self._show_object_not_found_warnings = True
        self._mobile = is_mobile_mode()

    def name(self) -> str:
        """Return DGI name."""
        return self._name

    def alias(self) -> str:
        """Return DGI alias."""
        return self._alias

    def create_app(self) -> "QtWidgets.QApplication":
        """Create an alternative Core.Application."""
        from pineboolib import application

        return application.PROJECT.app

    # Establece un lanzador alternativo al de la aplicación
    def alternativeMain(self, options: Any) -> Any:
        """Return alternative main."""
        return None

    def accept_file(self, name: str) -> bool:
        """Return True if file is accepted .False elsewhere."""
        return True

    def useDesktop(self) -> bool:
        """Return if desktop UI is used."""
        return self._desktop_enabled

    def setUseDesktop(self, val: bool) -> None:
        """Set if desktop UI is used."""
        self._desktop_enabled = val

    def localDesktop(
        self,
    ) -> bool:  # Indica si son ventanas locales o remotas a traves de algún parser
        """Return if is local desktop."""
        return self._local_desktop

    def setLocalDesktop(self, val: bool) -> None:
        """Set local desktop variable."""
        self._local_desktop = val

    def setUseMLDefault(self, val: bool) -> None:
        """Set if defaul main loader is used."""
        self._ml_default = val

    def useMLDefault(self) -> bool:
        """Return if main loaded is used."""
        return self._ml_default

    def setParameter(self, param: str) -> None:  # Se puede pasar un parametro al dgi
        """Set parameters to DGI."""
        pass

    def extraProjectInit(self) -> None:
        """Launch extra project init."""
        pass

    def showInitBanner(self) -> None:
        """Show init banner string."""

        print("")
        print("=============================================")
        print("                GDI_%s MODE               " % self._alias)
        print("=============================================")
        print("")
        print("")

    def mainForm(self) -> Any:
        """Return mainForm."""
        return QtCore.QObject()

    def interactiveGUI(self) -> str:
        """Return interactiveGUI name."""
        return "Pineboo"

    def show_object_not_found_warnings(self) -> bool:
        """Return if show warnings when objects not found."""
        return self._show_object_not_found_warnings

    def mobilePlatform(self) -> bool:
        """Return if run into a mobile platform."""
        return self._mobile

    def icon_size(self) -> QtCore.QSize:
        """Return default icon size."""

        size = QtCore.QSize(22, 22)
        # if self.mobilePlatform():
        #    size = QtCore.QSize(60, 60)

        return size

    def __getattr__(self, name: str) -> Optional["QtCore.QObject"]:
        """Return and object specified by name."""
        return self.resolveObject(self._name, name)

    def resolveObject(self, module_name: str, name: str) -> Optional["QtCore.QObject"]:
        """Return a DGI specific object."""
        cls = None
        mod_name_full = "pineboolib.plugins.dgi.dgi_%s.dgi_objects.%s" % (module_name, name.lower())
        try:
            # FIXME: Please, no.
            mod_ = import_module(mod_name_full)
            cls = getattr(mod_, name, None)
            LOGGER.trace("resolveObject: Loaded module %s", mod_name_full)
        except ModuleNotFoundError:
            LOGGER.trace("resolveObject: Module not found %s", mod_name_full)
        except Exception:
            LOGGER.exception("resolveObject: Unable to load module %s", mod_name_full)
        return cast(Optional[QtCore.QObject], cls)

    def use_alternative_credentials(self) -> bool:
        """Return True if use alternative authentication , False elsewhere."""
        return False

    def get_nameuser(self) -> str:
        """Return alternative user name."""
        return ""

    def debug(self, txt: str):
        """Show debug message."""
        LOGGER.info("---> %s" % txt)

    def exec(self):
        """Return default exec."""

        return 0



"""
Tests for loader.connection.
"""
import unittest
from pineboolib.loader.projectconfig import ProjectConfig
from pineboolib.loader.connection import config_dbconn  # , connect_to_db
from pineboolib.loader.options import parse_options

from pineboolib.loader.tests import fixture_path
from typing import TYPE_CHECKING

from unittest.mock import patch, Mock

if TYPE_CHECKING:
    import getpass  # noqa: F401 # pragma: no cover


class TestConfigDBConn(unittest.TestCase):
    """Test for config_dbconn function."""

    maxDiff = 200

    def test_connstring(self) -> None:
        """Test to provide a connstring."""
        # format: "user:passwd:driver_alias@host:port/database"
        options = parse_options(["pineboo", "--silentconn", "user:pass@127.0.0.1:5433/mydb"])
        cfg1 = config_dbconn(options)
        cfg2 = ProjectConfig(
            database="mydb",
            username="user",
            password="pass",
            host="127.0.0.1",
            port=5433,
            type="PostgreSQL (PSYCOPG2)",
        )
        self.assertEqual(cfg1, cfg2)

        options = parse_options(["pineboo", "--silentconn", "mydb"])
        cfg1 = config_dbconn(options)
        cfg2 = ProjectConfig(
            database="mydb",
            username="postgres",
            password="",
            host="127.0.0.1",
            port=5432,
            type="PostgreSQL (PSYCOPG2)",
        )
        self.assertEqual(cfg1, cfg2)

    def test_project(self) -> None:
        """Test to provide a project template."""
        options = parse_options(["pineboo", "--load", fixture_path("project_test1")])
        cfg1 = config_dbconn(options)
        cfg2 = ProjectConfig(database="mydb", type="SQLite3 (SQLITE3)")
        self.assertEqual(cfg1, cfg2)

    @patch("getpass.getpass")
    def test_project_passwd(self, mock_get_pass: Mock) -> None:
        """Test to provide a project template with password."""
        mock_get_pass.return_value = "myhardtoguesspassword"
        options = parse_options(["pineboo", "--load", fixture_path("project_test2")])
        cfg1 = config_dbconn(options)
        cfg2 = ProjectConfig(
            database="postgres_testdb",
            type="PostgreSQL (PSYCOPG2)",
            host="192.168.1.101",
            port=5432,
            username="postgres",
            password="postgrespassword",
            description="Postgres Test DB",
            project_password="myhardtoguesspassword",
        )
        self.assertEqual(cfg1, cfg2)
